"""Implements parameter scaling algorithms to update the parameters of a regularizer."""

from typing import Protocol, Tuple, Union

import numpy as np
import torch as th

LossType = Union[th.Tensor, float]


class LambdaUpdater(Protocol):
    """Protocol type for functions that update the regularizer parameter.

    A callable object that takes in the current lambda and the train and val loss, and
    returns the new lambda. This has been implemented as a protocol and not an ABC
    because a user might wish to provide their own implementation without having to
    inherit from the base class, e.g. by defining a function instead of a class.

    Note: if you implement `LambdaUpdater`, your implementation MUST be purely
    functional, i.e. side-effect free. The class structure should only be used
    to store constant hyperparameters. (Alternatively, closures can be used for that).
    """

    def __call__(self, lambda_, train_loss: LossType, val_loss: LossType) -> float:
        ...


class IntervalParamScaler(LambdaUpdater):
    """Scales the lambda of the regularizer by some constant factor.

    Lambda is scaled up if the ratio of the validation loss to the training loss
    is above the tolerable interval, and scaled down if the ratio is below the
    tolerable interval. Nothing happens if the ratio is within the tolerable
    interval.
    """

    def __init__(self, scaling_factor: float, tolerable_interval: Tuple[float, float]):
        """Initialize the interval parameter scaler.

        Args:
            scaling_factor: The factor by which to scale the lambda, a value in (0, 1).
            tolerable_interval: The interval within which the ratio of the validation
                loss to the training loss is considered acceptable. A tuple whose first
                element is at least 0 and the second element is greater than the first.

        Raises:
            ValueError: If the tolerable interval is not a tuple of length 2.
            ValueError: if the scaling factor is not in (0, 1).
            ValueError: if the tolerable interval is negative or not a proper interval.
        """
        eps = np.finfo(float).eps
        if not (eps < scaling_factor < 1 - eps):
            raise ValueError(
                "scaling_factor must be in (0, 1) within machine precision.",
            )
        if len(tolerable_interval) != 2:
            raise ValueError("tolerable_interval must be a tuple of length 2")
        if not (0 <= tolerable_interval[0] < tolerable_interval[1]):
            raise ValueError(
                "tolerable_interval must be a tuple whose first element "
                "is at least 0 and the second element is greater than "
                "the first",
            )

        self.scaling_factor = scaling_factor
        self.tolerable_interval = tolerable_interval

    def __call__(
        self,
        lambda_: float,
        train_loss: LossType,
        val_loss: LossType,
    ) -> float:
        """Scales the lambda of the regularizer by some constant factor.

        Lambda is scaled up if the ratio of the validation loss to the training loss
        is above the tolerable interval, and scaled down if the ratio is below the
        tolerable interval. Nothing happens if the ratio is within the tolerable
        interval.

        Args:
            lambda_: The current value of the lambda.
            train_loss: The loss on the training set.
            val_loss: The loss on the validation set.

        Returns:
            The new value of the lambda.

        Raises:
            ValueError: If the loss on the validation set is not a scalar.
            ValueError: if lambda_ is zero (will result in no scaling).
            ValueError: if lambda_ is not a float.
        """
        # check that the tensors val_loss and train_loss are both scalars
        if not (
            isinstance(val_loss, float)
            or (isinstance(val_loss, th.Tensor) and val_loss.dim() == 0)
        ):
            raise ValueError("val_loss must be a scalar")

        if not (
            isinstance(train_loss, float)
            or (isinstance(train_loss, th.Tensor) and train_loss.dim() == 0)
        ):
            raise ValueError("train_loss must be a scalar")

        if np.finfo(float).eps > abs(lambda_):
            raise ValueError(
                "lambda_ must not be zero. Make sure that you're not "
                "scaling the value of lambda down too quickly or passing an "
                "initial value of zero to the lambda parameter.",
            )
        elif lambda_ < 0:
            raise ValueError("lambda_ must be non-negative")
        if not isinstance(lambda_, float):
            raise ValueError("lambda_ must be a float")
        if train_loss < 0 or val_loss < 0:
            raise ValueError("losses must be non-negative for this updater")

        eps = np.finfo(float).eps
        if train_loss < eps and val_loss < eps:
            # 0/0 is undefined, so return the current lambda
            return lambda_
        elif train_loss < eps <= val_loss:
            # the ratio would be infinite
            return lambda_ * (1 + self.scaling_factor)

        val_to_train_ratio = val_loss / train_loss
        if val_to_train_ratio > self.tolerable_interval[1]:
            lambda_ *= 1 + self.scaling_factor
        elif val_to_train_ratio < self.tolerable_interval[0]:
            lambda_ *= 1 - self.scaling_factor
        return lambda_



"""Smoke tests for CLI programs in `imitation.scripts.*`.

Every test in this file should use `parallel=False` to turn off multiprocessing because
codecov might interact poorly with multiprocessing. The 'fast' named_config for each
experiment implicitly sets parallel=False.
"""

import collections
import filecmp
import os
import pathlib
import pickle
import platform
import shutil
import sys
import tempfile
from collections import Counter
from typing import Any, Dict, Generator, List, Mapping, Optional
from unittest import mock

import numpy as np
import pandas as pd
import pytest
import ray.tune as tune
import sacred
import sacred.utils
import stable_baselines3
import torch as th
from stable_baselines3.common import buffers
from stable_baselines3.her.her_replay_buffer import HerReplayBuffer

from imitation.data import serialize
from imitation.rewards import reward_nets
from imitation.scripts import (
    analyze,
    convert_trajs,
    eval_policy,
    parallel,
    train_adversarial,
    train_imitation,
    train_preference_comparisons,
    train_rl,
)
from imitation.util import networks, util

ALL_SCRIPTS_MODS = [
    analyze,
    eval_policy,
    parallel,
    train_adversarial,
    train_imitation,
    train_preference_comparisons,
    train_rl,
]

TEST_DATA_PATH = util.parse_path("tests/testdata")

if not TEST_DATA_PATH.exists():  # pragma: no cover
    raise RuntimeError(
        "Folder with test data has not been found. Make sure you are "
        "running tests relative to the base imitation project folder.",
    )

CARTPOLE_TEST_DATA_PATH = TEST_DATA_PATH / "expert_models/cartpole_0/"
CARTPOLE_TEST_ROLLOUT_PATH = CARTPOLE_TEST_DATA_PATH / "rollouts/final.npz"
CARTPOLE_TEST_POLICY_PATH = CARTPOLE_TEST_DATA_PATH / "policies/final"

PENDULUM_TEST_DATA_PATH = TEST_DATA_PATH / "expert_models/pendulum_0/"
PENDULUM_TEST_ROLLOUT_PATH = PENDULUM_TEST_DATA_PATH / "rollouts/final.npz"

PICKLE_FMT_ROLLOUT_TEST_DATA_PATH = TEST_DATA_PATH / "pickle_format_rollout.pkl"
NPZ_FMT_ROLLOUT_TEST_DATA_PATH = TEST_DATA_PATH / "npz_format_rollout.npz"


@pytest.fixture(autouse=True)
def sacred_capture_use_sys():
    """Set Sacred capture mode to "sys" because default "fd" option leads to error.

    See https://github.com/IDSIA/sacred/issues/289.

    Yields:
        None after setting capture mode; restores it after yield.
    """
    # TODO(shwang): Stop using non-default "sys" mode once the issue is fixed.
    temp = sacred.SETTINGS["CAPTURE_MODE"]
    sacred.SETTINGS.CAPTURE_MODE = "sys"
    yield
    sacred.SETTINGS.CAPTURE_MODE = temp


@pytest.mark.parametrize("script_mod", ALL_SCRIPTS_MODS)
def test_main_console(script_mod):
    """Smoke tests of main entry point for some cheap coverage."""
    argv = ["sacred-pytest-stub", "print_config"]
    with mock.patch.object(sys, "argv", argv):
        script_mod.main_console()


ALGO_FAST_CONFIGS = {
    "adversarial": [
        "environment.fast",
        "demonstrations.fast",
        "rl.fast",
        "policy_evaluation.fast",
        "fast",
    ],
    "eval_policy": ["environment.fast", "fast"],
    "imitation": [
        "environment.fast",
        "demonstrations.fast",
        "policy_evaluation.fast",
        "fast",
    ],
    "preference_comparison": [
        "environment.fast",
        "rl.fast",
        "policy_evaluation.fast",
        "fast",
    ],
    "rl": ["environment.fast", "rl.fast", "fast"],
}

RL_SAC_NAMED_CONFIGS = ["rl.sac", "policy.sac"]


@pytest.fixture(
    params=[
        "plain",
        "with_expert_trajectories",
        "warmstart",
        "with_checkpoints",
    ],
)
def preference_comparison_config(request):
    return dict(
        plain={},
        with_expert_trajectories={"trajectory_path": CARTPOLE_TEST_ROLLOUT_PATH},
        warmstart={
            # We're testing preference saving and disabling sampling here as well;
            # having yet another run just for those would be wasteful since they
            # don't interact with warm starting an agent.
            "save_preferences": True,
            "gatherer_kwargs": {"sample": False},
            "agent_path": CARTPOLE_TEST_POLICY_PATH,
            # FIXME(yawen): the policy we load was trained on 8 parallel environments
            #  and for some reason using it breaks if we use just 1 (like would be the
            #  default with the fast named_config)
            "environment": dict(num_vec=8),
        },
        with_checkpoints={
            # Test that we can save checkpoints
            "checkpoint_interval": 1,
        },
    )[request.param]


def test_train_preference_comparisons_main(tmpdir, preference_comparison_config):
    config_updates = dict(logging=dict(log_root=tmpdir))
    sacred.utils.recursive_update(config_updates, preference_comparison_config)
    run = train_preference_comparisons.train_preference_comparisons_ex.run(
        # Note: we have to use the cartpole and not the seals_cartpole config because
        #  the seals_cartpole config needs rollouts with a fixed horizon,
        #  and the saved trajectory rollouts are variable horizon.
        named_configs=["cartpole"] + ALGO_FAST_CONFIGS["preference_comparison"],
        config_updates=config_updates,
    )
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)


@pytest.mark.parametrize(
    "env_name",
    ["seals_cartpole", "mountain_car", "seals_mountain_car"],
)
def test_train_preference_comparisons_envs_no_crash(tmpdir, env_name):
    """Test envs specified in imitation.scripts.config.train_preference_comparisons."""
    config_updates = dict(logging=dict(log_root=tmpdir))
    run = train_preference_comparisons.train_preference_comparisons_ex.run(
        named_configs=[env_name] + ALGO_FAST_CONFIGS["preference_comparison"],
        config_updates=config_updates,
    )
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)


def test_train_preference_comparisons_sac(tmpdir):
    config_updates = dict(logging=dict(log_root=tmpdir))
    run = train_preference_comparisons.train_preference_comparisons_ex.run(
        # make sure rl.sac named_config is called after rl.fast to overwrite
        # rl_kwargs.batch_size to None
        named_configs=["pendulum"]
        + ALGO_FAST_CONFIGS["preference_comparison"]
        + RL_SAC_NAMED_CONFIGS,
        config_updates=config_updates,
    )
    assert run.config["rl"]["rl_cls"] is stable_baselines3.SAC
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)

    # Make sure rl.sac named_config is called after rl.fast to overwrite
    # rl_kwargs.batch_size to None
    with pytest.raises(Exception, match="set 'batch_size' at top-level"):
        train_preference_comparisons.train_preference_comparisons_ex.run(
            named_configs=["pendulum"]
            + RL_SAC_NAMED_CONFIGS
            + ALGO_FAST_CONFIGS["preference_comparison"],
            config_updates=config_updates,
        )


def test_train_preference_comparisons_sac_reward_relabel(tmpdir):
    def _run_reward_relabel_sac_preference_comparisons(buffer_cls):
        config_updates = dict(
            logging=dict(log_root=tmpdir),
            rl=dict(
                rl_kwargs=dict(
                    replay_buffer_class=buffer_cls,
                    replay_buffer_kwargs=dict(handle_timeout_termination=True),
                ),
            ),
        )
        run = train_preference_comparisons.train_preference_comparisons_ex.run(
            # make sure rl.sac named_config is called after rl.fast to overwrite
            # rl_kwargs.batch_size to None
            named_configs=["pendulum"]
            + ALGO_FAST_CONFIGS["preference_comparison"]
            + RL_SAC_NAMED_CONFIGS,
            config_updates=config_updates,
        )
        return run

    run = _run_reward_relabel_sac_preference_comparisons(buffers.ReplayBuffer)
    assert run.status == "COMPLETED"
    del run

    with pytest.raises(AssertionError, match="only ReplayBuffer is supported"):
        _run_reward_relabel_sac_preference_comparisons(buffers.DictReplayBuffer)
    with pytest.raises(AssertionError, match="only ReplayBuffer is supported"):
        _run_reward_relabel_sac_preference_comparisons(HerReplayBuffer)


@pytest.mark.parametrize(
    "named_configs",
    (
        [],
        ["reward.normalize_output_running"],
        ["reward.normalize_output_disable"],
    ),
)
def test_train_preference_comparisons_reward_named_config(tmpdir, named_configs):
    config_updates = dict(logging=dict(log_root=tmpdir))
    run = train_preference_comparisons.train_preference_comparisons_ex.run(
        named_configs=["seals_cartpole"]
        + ALGO_FAST_CONFIGS["preference_comparison"]
        + named_configs,
        config_updates=config_updates,
    )
    if "reward.normalize_output_running" in named_configs:
        assert run.config["reward"]["normalize_output_layer"] is networks.RunningNorm
    elif "reward.normalize_output_disable" in named_configs:
        assert run.config["reward"]["normalize_output_layer"] is None
    else:
        assert run.config["reward"]["normalize_output_layer"] is networks.RunningNorm
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)


def test_train_dagger_main(tmpdir):
    with pytest.warns(None) as record:
        run = train_imitation.train_imitation_ex.run(
            command_name="dagger",
            named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["imitation"],
            config_updates=dict(
                logging=dict(log_root=tmpdir),
                demonstrations=dict(path=CARTPOLE_TEST_ROLLOUT_PATH),
            ),
        )
    for warning in record:
        # PyTorch wants writeable arrays.
        # See https://github.com/HumanCompatibleAI/imitation/issues/219
        assert not (
            warning.category == UserWarning
            and "NumPy array is not writeable" in warning.message.args[0]
        )
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)


def test_train_dagger_warmstart(tmpdir):
    run = train_imitation.train_imitation_ex.run(
        command_name="dagger",
        named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["imitation"],
        config_updates=dict(
            logging=dict(log_root=tmpdir),
            demonstrations=dict(path=CARTPOLE_TEST_ROLLOUT_PATH),
        ),
    )
    assert run.status == "COMPLETED"

    log_dir = util.parse_path(run.config["logging"]["log_dir"])
    policy_path = log_dir / "scratch" / "policy-latest.pt"
    run_warmstart = train_imitation.train_imitation_ex.run(
        command_name="dagger",
        named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["imitation"],
        config_updates=dict(
            logging=dict(log_root=tmpdir),
            demonstrations=dict(path=CARTPOLE_TEST_ROLLOUT_PATH),
            bc=dict(agent_path=policy_path),
        ),
    )
    assert run_warmstart.status == "COMPLETED"
    assert isinstance(run_warmstart.result, dict)


def test_train_bc_main_with_none_demonstrations_raises_value_error(tmpdir):
    with pytest.raises(ValueError, match="n_expert_demos must be specified"):
        train_imitation.train_imitation_ex.run(
            command_name="bc",
            named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["imitation"],
            config_updates=dict(
                logging=dict(log_root=tmpdir),
                demonstrations=dict(n_expert_demos=None),
            ),
        )


def test_train_bc_main_with_demonstrations_from_huggingface(tmpdir):
    train_imitation.train_imitation_ex.run(
        command_name="bc",
        named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["imitation"],
        config_updates=dict(
            logging=dict(log_root=tmpdir),
            demonstrations=dict(
                source="huggingface",
                algo_name="ppo",
            ),
        ),
    )


def generate_imitation_config(tmpdir, request, command_name: str) -> Dict:
    environment_named_config = "seals_cartpole"

    if request.param == "expert_from_path":
        expert_config = dict(
            policy_type="ppo",
            loader_kwargs=dict(path=CARTPOLE_TEST_POLICY_PATH / "model.zip"),
        )
        # Note: we don't have a seals_cartpole expert in our testdata folder,
        # so we use the cartpole environment in this case.
        environment_named_config = "cartpole"
    elif request.param == "expert_from_huggingface":
        expert_config = dict(policy_type="ppo-huggingface")
    elif request.param == "random_expert":
        expert_config = dict(policy_type="random")
    elif request.param == "zero_expert":
        expert_config = dict(policy_type="zero")

    return dict(
        command_name=command_name,
        named_configs=[environment_named_config] + ALGO_FAST_CONFIGS["imitation"],
        config_updates=dict(
            logging=dict(log_root=tmpdir),
            expert=expert_config,
            demonstrations=dict(path=CARTPOLE_TEST_ROLLOUT_PATH),
        ),
    )


@pytest.fixture(
    params=[
        "expert_from_path",
        "expert_from_huggingface",
        "random_expert",
        "zero_expert",
    ],
)
def bc_config(tmpdir, request):
    return generate_imitation_config(tmpdir, request, "bc")


@pytest.fixture(
    params=[
        "expert_from_path",
        "expert_from_huggingface",
        "random_expert",
        "zero_expert",
    ],
)
def sqil_config(tmpdir, request):
    return generate_imitation_config(tmpdir, request, "sqil")


def test_train_bc_main(bc_config):
    run = train_imitation.train_imitation_ex.run(**bc_config)
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)


def test_train_bc_warmstart(tmpdir):
    run = train_imitation.train_imitation_ex.run(
        command_name="bc",
        named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["imitation"],
        config_updates=dict(
            logging=dict(log_root=tmpdir),
            demonstrations=dict(path=CARTPOLE_TEST_ROLLOUT_PATH),
            expert=dict(policy_type="ppo-huggingface"),
        ),
    )
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)

    policy_path = util.parse_path(run.config["logging"]["log_dir"]) / "final.th"
    run_warmstart = train_imitation.train_imitation_ex.run(
        command_name="bc",
        named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["imitation"],
        config_updates=dict(
            logging=dict(log_root=tmpdir),
            demonstrations=dict(path=CARTPOLE_TEST_ROLLOUT_PATH),
            bc=dict(agent_path=policy_path),
        ),
    )

    assert run_warmstart.status == "COMPLETED"
    assert isinstance(run_warmstart.result, dict)


def test_train_sqil_main(sqil_config):
    # NOTE: Having four different expert types as in bc might be overkill for sqil
    run = train_imitation.train_imitation_ex.run(**sqil_config)
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)


def test_train_sqil_main_with_demonstrations_from_huggingface(tmpdir):
    train_imitation.train_imitation_ex.run(
        command_name="sqil",
        named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["imitation"],
        config_updates=dict(
            logging=dict(log_root=tmpdir),
            demonstrations=dict(
                source="huggingface",
                algo_name="ppo",
            ),
        ),
    )


@pytest.fixture(params=["cold_start", "warm_start"])
def rl_train_ppo_config(request, tmpdir):
    config = dict(logging=dict(log_root=tmpdir))
    if request.param == "warm_start":
        # FIXME(yawen): the policy we load was trained on 8 parallel environments
        #  and for some reason using it breaks if we use just 1 (like would be the
        #  default with the fast named_config)
        config["agent_path"] = CARTPOLE_TEST_POLICY_PATH / "model.zip"
        config["environment"] = dict(num_vec=8)
    return config


def test_train_rl_main(tmpdir, rl_train_ppo_config):
    """Smoke test for imitation.scripts.train_rl."""
    config_updates = dict(logging=dict(log_root=tmpdir))
    sacred.utils.recursive_update(config_updates, rl_train_ppo_config)
    run = train_rl.train_rl_ex.run(
        named_configs=["cartpole"] + ALGO_FAST_CONFIGS["rl"],
        config_updates=config_updates,
    )
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)


def test_train_rl_wb_logging(tmpdir):
    """Smoke test for imitation.scripts.ingredients.logging.wandb_logging."""
    with pytest.raises(Exception, match="api_key not configured"):
        train_rl.train_rl_ex.run(
            named_configs=["cartpole"]
            + ALGO_FAST_CONFIGS["rl"]
            + ["logging.wandb_logging"],
            config_updates=dict(
                logging=dict(log_root=tmpdir),
            ),
        )


def test_train_rl_sac(tmpdir):
    run = train_rl.train_rl_ex.run(
        # make sure rl.sac named_config is called after rl.fast to overwrite
        # rl_kwargs.batch_size to None
        named_configs=["pendulum"] + ALGO_FAST_CONFIGS["rl"] + RL_SAC_NAMED_CONFIGS,
        config_updates=dict(
            logging=dict(log_root=tmpdir),
        ),
    )
    assert run.config["rl"]["rl_cls"] is stable_baselines3.SAC
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)


# check if platform is macos

EVAL_POLICY_CONFIGS: List[Dict] = [
    {"reward_type": "zero", "reward_path": "foobar"},
    {"explore_kwargs": {"switch_prob": 1.0, "random_prob": 0.1}},
    {"rollout_save_path": "{log_dir}/rollouts.npz"},
]

if platform.system() != "Darwin":
    EVAL_POLICY_CONFIGS.extend(
        [
            {"videos": True},
            {"videos": True, "video_kwargs": {"single_video": False}},
        ],
    )


@pytest.mark.parametrize("config", EVAL_POLICY_CONFIGS)
def test_eval_policy(config, tmpdir):
    """Smoke test for imitation.scripts.eval_policy."""
    config_updates = dict(logging=dict(log_root=tmpdir))
    config_updates.update(config)
    run = eval_policy.eval_policy_ex.run(
        config_updates=config_updates,
        named_configs=ALGO_FAST_CONFIGS["eval_policy"],
    )
    assert run.status == "COMPLETED"
    wrapped_reward = "reward_type" in config
    _check_rollout_stats(run.result, wrapped_reward)


def _check_rollout_stats(stats: dict, wrapped_reward: bool = True):
    """Common assertions for rollout_stats."""
    assert isinstance(stats, dict)
    assert "return_mean" in stats
    assert "monitor_return_mean" in stats
    if wrapped_reward:
        # If the reward is wrapped, then we expect the monitor return
        # to differ.
        assert stats.get("return_mean") != stats.get("monitor_return_mean")
    else:
        assert stats.get("return_mean") == stats.get("monitor_return_mean")


def _check_train_ex_result(result: dict):
    expert_stats = result.get("expert_stats")
    assert isinstance(expert_stats, dict)
    assert "return_mean" in expert_stats
    assert "monitor_return_mean" not in expert_stats

    imit_stats = result.get("imit_stats")
    assert isinstance(imit_stats, dict)
    _check_rollout_stats(imit_stats)


@pytest.mark.parametrize(
    "named_configs",
    (
        [],
        ["policy.normalize_running", "reward.normalize_input_running"],
        ["reward.normalize_input_disable"],
    ),
)
@pytest.mark.parametrize("command", ("airl", "gail"))
def test_train_adversarial(tmpdir, named_configs, command):
    """Smoke test for imitation.scripts.train_adversarial."""
    named_configs = (
        named_configs + ["seals_cartpole"] + ALGO_FAST_CONFIGS["adversarial"]
    )
    config_updates = {
        "logging": dict(log_root=tmpdir),
        "demonstrations": dict(path=CARTPOLE_TEST_ROLLOUT_PATH),
        # TensorBoard logs to get extra coverage
        "algorithm_kwargs": dict(init_tensorboard=True),
    }
    run = train_adversarial.train_adversarial_ex.run(
        command_name=command,
        named_configs=named_configs,
        config_updates=config_updates,
    )
    assert run.status == "COMPLETED"
    _check_train_ex_result(run.result)


@pytest.mark.parametrize("command", ("airl", "gail"))
def test_train_adversarial_warmstart(tmpdir, command):
    named_configs = ["cartpole"] + ALGO_FAST_CONFIGS["adversarial"]
    config_updates = {
        "logging": dict(log_root=tmpdir),
        "demonstrations": dict(path=CARTPOLE_TEST_ROLLOUT_PATH, source="local"),
    }
    run = train_adversarial.train_adversarial_ex.run(
        command_name=command,
        named_configs=named_configs,
        config_updates=config_updates,
    )

    log_dir = util.parse_path(run.config["logging"]["log_dir"])
    policy_path = log_dir / "checkpoints" / "final" / "gen_policy"

    run_warmstart = train_adversarial.train_adversarial_ex.run(
        command_name=command,
        named_configs=named_configs,
        config_updates={
            "agent_path": policy_path,
            **config_updates,
        },
    )

    assert run_warmstart.status == "COMPLETED"
    _check_train_ex_result(run_warmstart.result)


@pytest.mark.parametrize("command", ("airl", "gail"))
def test_train_adversarial_sac(tmpdir, command):
    """Smoke test for imitation.scripts.train_adversarial."""
    # Make sure rl.sac named_config is called after rl.fast to overwrite
    # rl_kwargs.batch_size to None
    named_configs = (
        ["pendulum"] + ALGO_FAST_CONFIGS["adversarial"] + RL_SAC_NAMED_CONFIGS
    )
    config_updates = {
        "logging": dict(log_root=tmpdir),
        "demonstrations": dict(path=PENDULUM_TEST_ROLLOUT_PATH),
    }
    run = train_adversarial.train_adversarial_ex.run(
        command_name=command,
        named_configs=named_configs,
        config_updates=config_updates,
    )
    assert run.config["rl"]["rl_cls"] is stable_baselines3.SAC
    assert run.status == "COMPLETED"
    _check_train_ex_result(run.result)


def test_train_adversarial_algorithm_value_error(tmpdir):
    """Error on bad algorithm arguments."""
    base_named_configs = ["cartpole"] + ALGO_FAST_CONFIGS["adversarial"]
    base_config_updates = collections.ChainMap(
        {
            "logging": dict(log_root=tmpdir),
            "demonstrations": dict(path=CARTPOLE_TEST_ROLLOUT_PATH, source="local"),
        },
    )

    with pytest.raises(TypeError, match="BAD_VALUE"):
        train_adversarial.train_adversarial_ex.run(
            command_name="gail",
            named_configs=base_named_configs,
            config_updates=base_config_updates.new_child(
                dict(algorithm_kwargs={"BAD_VALUE": "bar"}),
            ),
        )

    with pytest.raises(FileNotFoundError, match="BAD_VALUE"):
        train_adversarial.train_adversarial_ex.run(
            command_name="gail",
            named_configs=base_named_configs,
            config_updates=base_config_updates.new_child(
                {"demonstrations.path": "path/BAD_VALUE"},
            ),
        )

    n_traj = 1234567
    with pytest.raises(ValueError, match=f"{n_traj}"):
        train_adversarial.train_adversarial_ex.run(
            command_name="gail",
            named_configs=base_named_configs,
            config_updates=base_config_updates.new_child(
                {"demonstrations.n_expert_demos": n_traj},
            ),
        )


def test_transfer_learning(tmpdir: str) -> None:
    """Transfer learning smoke test.

    Saves a dummy AIRL test reward, then loads it for transfer learning.

    Args:
        tmpdir: Temporary directory to save results to.
    """
    tmpdir_path = util.parse_path(tmpdir)
    log_dir_train = tmpdir_path / "train"
    run = train_adversarial.train_adversarial_ex.run(
        command_name="airl",
        named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["adversarial"],
        config_updates=dict(
            logging=dict(log_dir=log_dir_train),
            demonstrations=dict(path=CARTPOLE_TEST_ROLLOUT_PATH),
        ),
    )
    assert run.status == "COMPLETED"
    _check_train_ex_result(run.result)

    _check_rollout_stats(run.result["imit_stats"])

    log_dir_data = tmpdir_path / "train_rl"
    reward_path = log_dir_train / "checkpoints" / "final" / "reward_test.pt"
    run = train_rl.train_rl_ex.run(
        named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["rl"],
        config_updates=dict(
            logging=dict(log_dir=log_dir_data),
            reward_type="RewardNet_unshaped",
            reward_path=reward_path,
        ),
    )
    assert run.status == "COMPLETED"
    _check_rollout_stats(run.result)


@pytest.mark.parametrize(
    "named_configs_dict",
    (
        dict(pc=[], rl=[]),
        dict(pc=["rl.sac", "policy.sac"], rl=["rl.sac", "policy.sac"]),
        dict(pc=["reward.reward_ensemble"], rl=[]),
    ),
)
def test_preference_comparisons_transfer_learning(
    tmpdir: str,
    named_configs_dict: Mapping[str, List[str]],
) -> None:
    """Transfer learning smoke test.

    Saves a preference comparisons ensemble reward, then loads it for transfer learning.

    Args:
        tmpdir: Temporary directory to save results to.
        named_configs_dict: Named configs for preference_comparisons and rl.
    """
    tmpdir_path = util.parse_path(tmpdir)

    log_dir_train = tmpdir_path / "train"
    run = train_preference_comparisons.train_preference_comparisons_ex.run(
        named_configs=["pendulum"]
        + ALGO_FAST_CONFIGS["preference_comparison"]
        + named_configs_dict["pc"],
        config_updates=dict(logging=dict(log_dir=log_dir_train)),
    )
    assert run.status == "COMPLETED"

    if "reward.reward_ensemble" in named_configs_dict["pc"]:
        assert run.config["reward"]["net_cls"] is reward_nets.RewardEnsemble
        assert run.config["reward"]["add_std_alpha"] == 0.0
        reward_type = "RewardNet_std_added"
        load_reward_kwargs = {"alpha": -1}
    else:
        reward_type = "RewardNet_unnormalized"
        load_reward_kwargs = {}

    log_dir_data = tmpdir_path / "train_rl"
    reward_path = log_dir_train / "checkpoints" / "final" / "reward_net.pt"
    agent_path = log_dir_train / "checkpoints" / "final" / "policy"
    run = train_rl.train_rl_ex.run(
        named_configs=["pendulum"] + ALGO_FAST_CONFIGS["rl"] + named_configs_dict["rl"],
        config_updates=dict(
            logging=dict(log_dir=log_dir_data),
            reward_type=reward_type,
            reward_path=reward_path,
            load_reward_kwargs=load_reward_kwargs,
            agent_path=agent_path,
        ),
    )
    assert run.status == "COMPLETED"
    _check_rollout_stats(run.result)


def test_train_rl_double_normalization(tmpdir: str, rng):
    venv = util.make_vec_env(
        "CartPole-v1",
        n_envs=1,
        parallel=False,
        rng=rng,
    )
    basic_reward_net = reward_nets.BasicRewardNet(
        venv.observation_space,
        venv.action_space,
    )
    net = reward_nets.NormalizedRewardNet(basic_reward_net, networks.RunningNorm)
    tmppath = os.path.join(tmpdir, "reward.pt")
    th.save(net, tmppath)

    log_dir_data = os.path.join(tmpdir, "train_rl")
    with pytest.warns(
        RuntimeWarning,
        match=r"Applying normalization to already normalized reward function",
    ):
        train_rl.train_rl_ex.run(
            named_configs=["cartpole"] + ALGO_FAST_CONFIGS["rl"],
            config_updates=dict(
                logging=dict(log_dir=log_dir_data),
                reward_type="RewardNet_normalized",
                normalize_reward=True,
                reward_path=tmppath,
            ),
        )


def test_train_rl_cnn_policy(tmpdir: str, rng):
    venv = util.make_vec_env(
        "AsteroidsNoFrameskip-v4",
        n_envs=1,
        parallel=False,
        rng=rng,
    )
    net = reward_nets.CnnRewardNet(venv.observation_space, venv.action_space)
    tmppath = os.path.join(tmpdir, "reward.pt")
    th.save(net, tmppath)

    log_dir_data = os.path.join(tmpdir, "train_rl")
    run = train_rl.train_rl_ex.run(
        named_configs=["policy.cnn_policy"] + ALGO_FAST_CONFIGS["rl"],
        config_updates=dict(
            environment=dict(gym_id="AsteroidsNoFrameskip-v4"),
            logging=dict(log_dir=log_dir_data),
            reward_path=tmppath,
        ),
    )
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)


PARALLEL_CONFIG_UPDATES = [
    dict(
        sacred_ex_name="train_rl",
        base_named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["rl"],
        repeat=2,
        search_space={
            "config_updates": {
                "rl": {"rl_kwargs": {"learning_rate": tune.choice([3e-4, 1e-4])}},
            },
            "meta_info": {"asdf": "I exist for coverage purposes"},
        },
    ),
    dict(
        sacred_ex_name="train_adversarial",
        base_named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["adversarial"],
        base_config_updates={
            # Need absolute path because raylet runs in different working directory.
            "demonstrations.path": CARTPOLE_TEST_ROLLOUT_PATH.absolute(),
        },
        search_space={
            "command_name": "airl",
            "config_updates": {"total_timesteps": tune.choice([5, 10])},
        },
    ),
]

PARALLEL_CONFIG_LOW_RESOURCE = {
    # CI server only has 2 cores.
    "init_kwargs": {"num_cpus": 2},
    # Memory is low enough we only want to run one job at a time.
    "resources_per_trial": {"cpu": 2},
}


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Ray is buggy on windows.")
@pytest.mark.parametrize("config_updates", PARALLEL_CONFIG_UPDATES)
def test_parallel(config_updates, tmpdir):
    """Hyperparam tuning smoke test."""
    # CI server only has 2 cores
    config_updates = dict(config_updates)
    config_updates.update(PARALLEL_CONFIG_LOW_RESOURCE)
    config_updates.setdefault("base_config_updates", {})["logging.log_root"] = tmpdir
    run = parallel.parallel_ex.run(config_updates=config_updates)
    assert run.status == "COMPLETED"


def test_parallel_arg_errors(tmpdir):
    """Error on bad algorithm arguments."""
    config_updates = dict(PARALLEL_CONFIG_LOW_RESOURCE)
    config_updates.setdefault("base_config_updates", {})["logging.log_root"] = tmpdir
    config_updates = collections.ChainMap(config_updates)

    with pytest.raises(TypeError, match="Sequence"):
        parallel.parallel_ex.run(
            config_updates=config_updates.new_child(dict(base_named_configs={})),
        )

    with pytest.raises(TypeError, match="Mapping"):
        parallel.parallel_ex.run(
            config_updates=config_updates.new_child(dict(base_config_updates=())),
        )

    with pytest.raises(TypeError, match="Sequence"):
        parallel.parallel_ex.run(
            config_updates=config_updates.new_child(
                dict(search_space={"named_configs": {}}),
            ),
        )

    with pytest.raises(TypeError, match="Mapping"):
        parallel.parallel_ex.run(
            config_updates=config_updates.new_child(
                dict(search_space={"config_updates": ()}),
            ),
        )


def _generate_test_rollouts(tmpdir: str, env_named_config: str) -> pathlib.Path:
    tmpdir_path = util.parse_path(tmpdir)
    train_rl.train_rl_ex.run(
        named_configs=[env_named_config] + ALGO_FAST_CONFIGS["rl"],
        config_updates=dict(
            logging=dict(log_dir=tmpdir),
        ),
    )
    path = tmpdir_path / "rollouts/final.npz"
    return path.absolute()


def test_parallel_train_adversarial_custom_env(tmpdir):
    if os.name == "nt":  # pragma: no cover
        pytest.skip(
            "`ray.init()` times out when this test runs concurrently with other "
            "test_parallel tests on Windows (e.g., `pytest -n auto -k test_parallel`)",
        )

    env_named_config = "pendulum"
    path = _generate_test_rollouts(tmpdir, env_named_config)

    config_updates = dict(
        sacred_ex_name="train_adversarial",
        repeat=2,
        base_named_configs=[env_named_config] + ALGO_FAST_CONFIGS["adversarial"],
        base_config_updates=dict(
            logging=dict(log_root=tmpdir),
            demonstrations=dict(path=path),
        ),
        # specifying repeat=2 uses the optuna search algorithm which
        # requires the search space to be non-empty. So we provide
        # the command name using tune.choice.
        search_space=dict(command_name=tune.choice(["gail"])),
    )
    config_updates.update(PARALLEL_CONFIG_LOW_RESOURCE)
    run = parallel.parallel_ex.run(config_updates=config_updates)
    assert run.status == "COMPLETED"


def _run_train_adv_for_test_analyze_imit(run_name, sacred_logs_dir, log_dir):
    run = train_adversarial.train_adversarial_ex.run(
        command_name="gail",
        named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["adversarial"],
        config_updates=dict(
            logging=dict(log_root=log_dir),
            demonstrations=dict(path=CARTPOLE_TEST_ROLLOUT_PATH),
            checkpoint_interval=-1,
        ),
        options={"--name": run_name, "--file_storage": sacred_logs_dir},
    )
    return run


def _run_train_bc_for_test_analyze_imit(run_name, sacred_logs_dir, log_dir):
    run = train_imitation.train_imitation_ex.run(
        command_name="bc",
        named_configs=["seals_cartpole"] + ALGO_FAST_CONFIGS["imitation"],
        config_updates=dict(
            logging=dict(log_dir=log_dir),
            demonstrations=dict(path=CARTPOLE_TEST_ROLLOUT_PATH),
        ),
        options={"--name": run_name, "--file_storage": sacred_logs_dir},
    )
    return run


@pytest.mark.parametrize("run_names", ([], list("adab")))
@pytest.mark.parametrize(
    "run_sacred_fn",
    (
        _run_train_adv_for_test_analyze_imit,
        _run_train_bc_for_test_analyze_imit,
    ),
)
def test_analyze_imitation(tmpdir: str, run_names: List[str], run_sacred_fn):
    sacred_logs_dir = tmpdir_path = util.parse_path(tmpdir)

    # Generate sacred logs (other logs are put in separate tmpdir for deletion).
    for run_name in run_names:
        with tempfile.TemporaryDirectory(prefix="junk") as log_dir:
            run = run_sacred_fn(run_name, sacred_logs_dir, log_dir)
            assert run.status == "COMPLETED"

    # Check that analyze script finds the correct number of logs.
    def check(run_name: Optional[str], count: int, table_verbosity: int = 1) -> None:
        run = analyze.analysis_ex.run(
            command_name="analyze_imitation",
            config_updates=dict(
                source_dirs=[sacred_logs_dir],
                env_name="seals/CartPole-v0",
                run_name=run_name,
                csv_output_path=tmpdir_path / "analysis.csv",
                tex_output_path=tmpdir_path / "analysis.tex",
                print_table=True,
                table_verbosity=table_verbosity,
            ),
        )
        assert run.status == "COMPLETED"
        df = pd.DataFrame(run.result)
        assert df.shape[0] == count

    for run_name, count in Counter(run_names).items():
        check(run_name, count)

    check(None, len(run_names), table_verbosity=3)  # Check total number of logs.


def test_analyze_gather_tb(tmpdir: str):
    if os.name == "nt":  # pragma: no cover
        pytest.skip("gather_tb uses symlinks: not supported by Windows")
    num_runs = 2
    config_updates: Dict[str, Any] = dict(
        tune_run_kwargs=dict(local_dir=tmpdir),
        run_name="test",
    )
    config_updates.update(PARALLEL_CONFIG_LOW_RESOURCE)
    config_updates.update(num_samples=num_runs)
    parallel_run = parallel.parallel_ex.run(
        named_configs=["generate_test_data"],
        config_updates=config_updates,
    )
    assert parallel_run.status == "COMPLETED"

    run = analyze.analysis_ex.run(
        command_name="gather_tb_directories",
        config_updates=dict(
            source_dirs=[tmpdir],
        ),
    )
    assert run.status == "COMPLETED"
    assert isinstance(run.result, dict)
    assert run.result["n_tb_dirs"] == num_runs


def test_pickle_fmt_rollout_test_data_is_pickle():
    # WHEN
    with open(PICKLE_FMT_ROLLOUT_TEST_DATA_PATH, "rb") as f:
        pickle.load(f)

    # THEN
    # No exception raised


def test_npz_fmt_rollout_test_data_is_npz():
    # WHEN
    np.load(NPZ_FMT_ROLLOUT_TEST_DATA_PATH, allow_pickle=False)

    # THEN
    # No exception raised


@pytest.fixture(
    params=[PICKLE_FMT_ROLLOUT_TEST_DATA_PATH, NPZ_FMT_ROLLOUT_TEST_DATA_PATH],
    ids=["pickle", "npz"],
)
def old_rollouts_file_in_tmp(tmpdir, request) -> Generator[str, None, None]:
    old_rollouts_file = request.param
    shutil.copy(old_rollouts_file, tmpdir)
    yield os.path.join(tmpdir, os.path.basename(old_rollouts_file))


@pytest.fixture()
def new_rollouts_file_in_tmp(tmpdir) -> Generator[pathlib.Path, None, None]:
    # Note: here we just convert some old rollouts to the new format, so we don't need
    #  to keep trajectories in the new format in the testdata folder
    npz_rollouts_in_tmp = shutil.copy(NPZ_FMT_ROLLOUT_TEST_DATA_PATH, tmpdir)
    yield convert_trajs.update_traj_file_in_place(npz_rollouts_in_tmp)


def test_converted_trajectories_equal_original(old_rollouts_file_in_tmp: str):
    # GIVEN
    old_trajs = serialize.load(old_rollouts_file_in_tmp)

    # WHEN
    converted_path = convert_trajs.update_traj_file_in_place(old_rollouts_file_in_tmp)

    # THEN
    converted_trajs = serialize.load(converted_path)

    assert len(old_trajs) == len(converted_trajs)
    for t_old, t_converted in zip(old_trajs, converted_trajs):
        assert t_old == t_converted


def test_convert_trajs_from_current_format_is_idempotent(
    new_rollouts_file_in_tmp: pathlib.Path,
):
    # GIVEN
    original_path = new_rollouts_file_in_tmp.with_suffix(".orig")

    # WHEN
    shutil.copytree(new_rollouts_file_in_tmp, original_path)
    converted_path = convert_trajs.update_traj_file_in_place(new_rollouts_file_in_tmp)

    # THEN
    assert (
        filecmp.dircmp(converted_path, original_path).diff_files == []
    ), "convert_trajs not idempotent"



"""Tests for imitation.scripts.ingredients.reward."""
from typing import Any, Mapping

import pytest

from imitation.rewards import reward_nets
from imitation.scripts.ingredients import reward
from imitation.util import networks


@pytest.fixture
def member_config() -> Mapping[str, Any]:
    return {
        "net_cls": reward_nets.BasicRewardNet,
        "net_kwargs": {},
        "normalize_output_layer": None,
    }


def test_make_reward_ensemble(member_config, cartpole_venv):
    reward_net = reward.make_reward_net(
        venv=cartpole_venv,
        net_cls=reward_nets.RewardEnsemble,
        add_std_alpha=None,
        net_kwargs={},
        normalize_output_layer=None,
        ensemble_size=3,
        ensemble_member_config=member_config,
    )
    assert isinstance(reward_net, reward_nets.RewardEnsemble)

    reward_net = reward.make_reward_net(
        venv=cartpole_venv,
        net_cls=reward_nets.RewardEnsemble,
        add_std_alpha=0,
        net_kwargs={},
        normalize_output_layer=None,
        ensemble_size=3,
        ensemble_member_config=member_config,
    )
    assert isinstance(reward_net, reward_nets.AddSTDRewardWrapper)


def test_make_reward_errors(member_config, cartpole_venv):
    with pytest.raises(ValueError, match=r"Must specify ensemble_size."):
        reward.make_reward_net(
            venv=cartpole_venv,
            net_cls=reward_nets.RewardEnsemble,
            add_std_alpha=None,
            net_kwargs={},
            normalize_output_layer=None,
            ensemble_size=None,
            ensemble_member_config=member_config,
        )

    with pytest.raises(ValueError, match=r"Must specify ensemble_member_config."):
        reward.make_reward_net(
            venv=cartpole_venv,
            net_cls=reward_nets.RewardEnsemble,
            add_std_alpha=None,
            net_kwargs={},
            normalize_output_layer=None,
            ensemble_size=5,
            ensemble_member_config=None,
        )

    with pytest.raises(
        ValueError,
        match=r"Output normalization not supported on RewardEnsembles.",
    ):
        reward.make_reward_net(
            venv=cartpole_venv,
            net_cls=reward_nets.RewardEnsemble,
            add_std_alpha=None,
            net_kwargs={},
            normalize_output_layer=networks.RunningNorm,
            ensemble_size=5,
            ensemble_member_config=member_config,
        )



"""Converts a directory of Sacred output to a CSV file."""
import pathlib
import sys

from imitation.util.sacred_file_parsing import find_sacred_runs


def main(path: pathlib.Path, only_completed_runs: bool = True):
    if not path.exists():
        raise NotADirectoryError(f"Path {path} does not exist.")

    # Print header
    if only_completed_runs:
        print("algo, env, score, expert_score")
    else:
        print("algo, env, score, expert_score, status")

    # Print data
    for config, run in find_sacred_runs(path, only_completed_runs):
        algo = run["command"]
        env = config["environment"]["gym_id"]
        score = run["result"]["imit_stats"]["monitor_return_mean"]
        expert_score = run["result"]["expert_stats"]["monitor_return_mean"]

        if only_completed_runs:
            print(f"{algo}, {env}, {score}, {expert_score}")
        else:
            status = run["status"]
            print(f"{algo}, {env}, {score}, {expert_score}, {status}")


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <path to sacred output directory>")
        sys.exit(1)

    main(pathlib.Path(sys.argv[1]))



"""Commands to analyze experimental results."""

import itertools
import json
import logging
import os
import pathlib
import tempfile
import warnings
from typing import Any, Callable, Iterable, List, Mapping, Optional, Sequence, Set

import pandas as pd
from sacred.observers import FileStorageObserver

import imitation.util.sacred as sacred_util
from imitation.scripts.config.analyze import analysis_ex
from imitation.util import util
from imitation.util.sacred import dict_get_nested as get


@analysis_ex.capture
def _gather_sacred_dicts(
    source_dirs: Sequence[str],
    run_name: Optional[str],
    env_name: Optional[str],
    skip_failed_runs: bool,
) -> List[sacred_util.SacredDicts]:
    """Helper function for parsing and selecting Sacred experiment JSON files.

    Args:
        source_dirs: A directory containing Sacred FileObserver subdirectories
            associated with the `train_adversarial` Sacred script. Behavior is
            undefined if there are Sacred subdirectories associated with other
            scripts. (Captured argument)
        run_name: If provided, then only analyze results from Sacred directories
            associated with this run name. `run_name` is compared against the
            "experiment.name" key in `run.json`. (Captured argument)
        env_name: If provided, then only analyze results from Sacred directories
            associated with this Gym environment ID.
        skip_failed_runs: If True, then filter out runs where the status is FAILED.
            (Captured argument)

    Returns:
        A list of `SacredDicts` corresponding to the selected Sacred directories.
    """
    # e.g. chain.from_iterable([["pathone", "pathtwo"], [], ["paththree"]]) =>
    # ("pathone", "pathtwo", "paththree")
    sacred_dirs = itertools.chain.from_iterable(
        sacred_util.filter_subdirs(util.parse_path(source_dir))
        for source_dir in source_dirs
    )
    sacred_dicts_list = []

    for sacred_dir in sacred_dirs:
        try:
            sacred_dicts_list.append(sacred_util.SacredDicts.load_from_dir(sacred_dir))
        except json.JSONDecodeError:
            warnings.warn(f"Invalid JSON file in {sacred_dir}", RuntimeWarning)

    sacred_dicts: Iterable = sacred_dicts_list
    if run_name is not None:
        sacred_dicts = filter(
            lambda sd: get(sd.run, "experiment.name") == run_name,
            sacred_dicts,
        )

    if env_name is not None:
        sacred_dicts = filter(
            lambda sd: get(sd.config, "environment.gym_id") == env_name,
            sacred_dicts,
        )

    if skip_failed_runs:
        sacred_dicts = filter(
            lambda sd: get(sd.run, "status") != "FAILED",
            sacred_dicts,
        )

    return list(sacred_dicts)


@analysis_ex.command
def gather_tb_directories() -> dict:
    """Gather Tensorboard directories from a `parallel_ex` run.

    The directories are copied to a unique directory in `/tmp/analysis_tb/` under
    subdirectories matching the Tensorboard events' Ray Tune trial names.

    This function calls the helper `_gather_sacred_dicts`, which captures its arguments
    automatically via Sacred. Provide those arguments to select which Sacred
    results to parse.

    Returns:
        A dict with two keys. "gather_dir" (str) is a path to a /tmp/ directory
        containing all the TensorBoard runs filtered from `source_dir`.
        "n_tb_dirs" (int) is the number of TensorBoard directories that were filtered.

    Raises:
        OSError: If the symlink cannot be created.
    """
    tb_analysis_dir = pathlib.Path("/tmp/analysis_tb")
    tb_analysis_dir.mkdir(exist_ok=True)
    tmp_dir = pathlib.Path(tempfile.mkdtemp(dir=tb_analysis_dir))

    tb_dirs_count = 0
    for sd in _gather_sacred_dicts():
        # Expecting a path like "~/ray_results/{run_name}/sacred/1".
        # Want to search for all Tensorboard dirs inside
        # "~/ray_results/{run_name}".
        run_dir = sd.sacred_dir.parent.parent
        run_name = run_dir.name

        # log is what we use as subdirectory in new code.
        # rl, tb, sb_tb all appear in old versions.
        for basename in ["log", "rl", "tb", "sb_tb"]:
            tb_src_dirs = tuple(
                sacred_util.filter_subdirs(
                    run_dir,
                    lambda path: path.name == basename,
                ),
            )
            if tb_src_dirs:
                assert len(tb_src_dirs) == 1, "expect at most one TB dir of each type"
                tb_src_dir = tb_src_dirs[0]

                symlinks_dir = tmp_dir / basename
                symlinks_dir.mkdir(exist_ok=True)

                tb_symlink = symlinks_dir / run_name
                try:
                    tb_symlink.symlink_to(tb_src_dir)
                except OSError as e:
                    if os.name == "nt":  # Windows
                        msg = (
                            "Exception occurred while creating symlink. "
                            "Please ensure that Developer mode is enabled."
                        )
                        raise OSError(msg) from e
                    else:
                        raise e
                tb_dirs_count += 1

    logging.info(f"Symlinked {tb_dirs_count} TensorBoard dirs to {tmp_dir}.")
    logging.info(f"Start Tensorboard with `tensorboard --logdir {tmp_dir}`.")
    return {"n_tb_dirs": tb_dirs_count, "gather_dir": tmp_dir}


def _get_exp_command(sd: sacred_util.SacredDicts) -> str:
    return str(sd.run.get("command"))


def _get_algo_name(sd: sacred_util.SacredDicts) -> str:
    exp_command = _get_exp_command(sd)

    if exp_command == "gail":
        return "GAIL"
    elif exp_command == "airl":
        return "AIRL"
    elif exp_command == "train_bc":
        return "BC"
    elif exp_command == "train_dagger":
        return "DAgger"
    else:
        return f"??exp_command={exp_command}"


def _return_summaries(sd: sacred_util.SacredDicts) -> dict:
    imit_stats = get(sd.run, "result.imit_stats")
    expert_stats = get(sd.run, "result.expert_stats")

    expert_return_summary = None
    if expert_stats is not None:
        expert_return_summary = _make_return_summary(expert_stats)

    imit_return_summary = None
    if imit_stats is not None:
        imit_return_summary = _make_return_summary(imit_stats, "monitor_")

    if imit_stats is not None and expert_stats is not None:
        # Assuming here that `result.imit_stats` and `result.expert_stats` are
        # formatted correctly.
        imit_expert_ratio = (
            imit_stats["monitor_return_mean"] / expert_stats["return_mean"]
        )
    else:
        imit_expert_ratio = None

    return dict(
        expert_stats=expert_stats,
        imit_stats=imit_stats,
        expert_return_summary=expert_return_summary,
        imit_return_summary=imit_return_summary,
        imit_expert_ratio=imit_expert_ratio,
    )


sd_to_table_entry_type = Mapping[str, Callable[[sacred_util.SacredDicts], Any]]

# This dict maps column names to functions that get table entries, given the
# row's unique SacredDicts object.
table_entry_fns: sd_to_table_entry_type = {
    "status": lambda sd: get(sd.run, "status"),
    "exp_command": _get_exp_command,
    "algo": _get_algo_name,
    "env_name": lambda sd: get(sd.config, "environment.gym_id"),
    "n_expert_demos": lambda sd: get(sd.config, "demonstrations.n_expert_demos"),
    "run_name": lambda sd: get(sd.run, "experiment.name"),
    "expert_return_summary": lambda sd: _return_summaries(sd)["expert_return_summary"],
    "imit_return_summary": lambda sd: _return_summaries(sd)["imit_return_summary"],
    "imit_expert_ratio": lambda sd: _return_summaries(sd)["imit_expert_ratio"],
}

# If `verbosity` is at least the length of this list, then we use all table_entry_fns
# as columns of table.
# Otherwise, use only the subset at index `verbosity`. The subset of columns is
# still arranged in the same order as in the `table_entry_fns` dict.
table_verbosity_mapping: List[Set[str]] = []

# verbosity 0
table_verbosity_mapping.append(
    {
        "algo",
        "env_name",
        "expert_return_summary",
        "imit_return_summary",
    },
)

# verbosity 1
table_verbosity_mapping.append(table_verbosity_mapping[-1] | {"n_expert_demos"})

# verbosity 2
table_verbosity_mapping.append(
    table_verbosity_mapping[-1]
    | {"status", "imit_expert_ratio", "exp_command", "run_name"},
)


def _get_table_entry_fns_subset(table_verbosity: int) -> sd_to_table_entry_type:
    assert table_verbosity >= 0
    if table_verbosity >= len(table_verbosity_mapping):
        return table_entry_fns
    else:
        keys_subset = table_verbosity_mapping[table_verbosity]
        return {k: v for k, v in table_entry_fns.items() if k in keys_subset}


@analysis_ex.command
def analyze_imitation(
    csv_output_path: Optional[str],
    tex_output_path: Optional[str],
    print_table: bool,
    table_verbosity: int,
) -> pd.DataFrame:
    """Parse Sacred logs and generate a DataFrame for imitation learning results.

    This function calls the helper `_gather_sacred_dicts`, which captures its arguments
    automatically via Sacred. Provide those arguments to select which Sacred
    results to parse.

    Args:
        csv_output_path: If provided, then save a CSV output file to this path.
        tex_output_path: If provided, then save a LaTeX-format table to this path.
        print_table: If True, then print the dataframe to stdout.
        table_verbosity: Increasing levels of verbosity, from 0 to 3, increase the
            number of columns in the table. Level 3 prints all of the columns available.

    Returns:
        The DataFrame generated from the Sacred logs.
    """
    # Get column names for which we have get value using make_entry_fn
    # These are same across Level 2 & 3. In Level 3, we additionally add remaining
    #  config columns.
    table_entry_fns_subset = _get_table_entry_fns_subset(min(table_verbosity, 2))

    output_table = pd.DataFrame()
    for sd in _gather_sacred_dicts():
        if table_verbosity == 3:
            # gets all config columns
            row = pd.json_normalize(sd.config)
        else:
            # create an empty dataframe with a single row
            row = pd.DataFrame(index=[0])

        for col_name, make_entry_fn in table_entry_fns_subset.items():
            row[col_name] = make_entry_fn(sd)

        output_table = pd.concat([output_table, row])

    if len(output_table) > 0:
        output_table.sort_values(by=["algo", "env_name"], inplace=True)

    display_options: Mapping[str, Any] = dict(index=False)
    if csv_output_path is not None:
        output_table.to_csv(csv_output_path, **display_options)
        print(f"Wrote CSV file to {csv_output_path}")
    if tex_output_path is not None:
        s: str = output_table.to_latex(**display_options)
        with open(tex_output_path, "w") as f:
            f.write(s)
        print(f"Wrote TeX file to {tex_output_path}")

    if print_table:
        print(output_table.to_string(**display_options))
    return output_table


def _make_return_summary(stats: dict, prefix="") -> str:
    return "{:3g} ± {:3g} (n={})".format(
        stats[f"{prefix}return_mean"],
        stats[f"{prefix}return_std"],
        stats["n_traj"],
    )


def main_console():
    observer_path = pathlib.Path.cwd() / "output" / "sacred" / "analyze"
    observer = FileStorageObserver(observer_path)
    analysis_ex.observers.append(observer)
    analysis_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()



"""Tests for imitation.algorithms.base."""

import numpy as np
import pytest
import torch as th

from imitation.algorithms import base
from imitation.data import types


def test_check_fixed_horizon(custom_logger):
    """Tests check for fixed horizon catches trajectories of varying lengths."""
    algo = base.BaseImitationAlgorithm(custom_logger=custom_logger)
    algo._check_fixed_horizon([])
    assert algo._horizon is None
    algo._check_fixed_horizon([5])
    assert algo._horizon == 5
    algo._check_fixed_horizon([5])
    algo._check_fixed_horizon([])
    algo._check_fixed_horizon([5, 5, 5])

    with pytest.raises(ValueError, match="Episodes of different length.*"):
        algo._check_fixed_horizon([4])
    with pytest.raises(ValueError, match="Episodes of different length.*"):
        algo._check_fixed_horizon([6])
    with pytest.raises(ValueError, match="Episodes of different length.*"):
        algo._check_fixed_horizon([1])
    assert algo._horizon == 5


def test_check_fixed_horizon_flag(custom_logger):
    """Tests check for fixed horizon ignores variable horizon with allow flag."""
    algo = base.BaseImitationAlgorithm(
        custom_logger=custom_logger,
        allow_variable_horizon=True,
    )
    algo._check_fixed_horizon([5])
    algo._check_fixed_horizon([42])
    algo._check_fixed_horizon([5, 42])
    assert algo._horizon is None


def _make_and_iterate_loader(*args, **kwargs):
    # our pytype version doesn't understand optional arguments in TypedDict
    # this is fixed in 2023.04.11, but we require 2022.7.26
    # See https://github.com/google/pytype/issues/1195
    loader = base.make_data_loader(*args, **kwargs)  # pytype: disable=wrong-arg-types
    for batch in loader:
        pass


def test_make_data_loader_batch_size():
    """Tests data loader performs batch size validation."""
    for batch_size in [0, -1, -42]:
        with pytest.raises(ValueError, match=".*must be positive"):
            base.make_data_loader([], batch_size=batch_size)

    # batch size = 5
    batch_iterable = [{"obs": np.zeros((5, 2)), "acts": np.zeros((5, 1))}]
    for wrong_batch_size in [4, 6, 42]:
        with pytest.raises(ValueError, match="Expected batch size.*"):
            _make_and_iterate_loader(batch_iterable, batch_size=wrong_batch_size)
    _make_and_iterate_loader(batch_iterable, batch_size=5)

    batch_iterable2 = [{"obs": np.zeros((5, 2)), "acts": np.zeros((4, 1))}]
    with pytest.raises(ValueError, match="Expected batch size.*"):
        _make_and_iterate_loader(batch_iterable2, batch_size=5)

    batch_iterable3 = [
        {"obs": np.zeros((5, 2)), "acts": np.zeros((5, 1))},
        {"obs": np.zeros((6, 2)), "acts": np.zeros((5, 1))},
    ]
    with pytest.raises(ValueError, match="Expected batch size.*"):
        _make_and_iterate_loader(batch_iterable3, batch_size=5)

    trans = types.TransitionsMinimal(
        obs=np.zeros((5, 2)),
        acts=np.zeros((5, 1)),
        infos=np.array([{}] * 5),
    )
    for smaller_bs in range(1, 6):
        base.make_data_loader(trans, batch_size=smaller_bs)
    for larger_bs in [6, 7, 42]:
        with pytest.raises(ValueError, match=".* smaller than batch size.*"):
            base.make_data_loader(trans, batch_size=larger_bs)


def test_make_data_loader():
    """Tests data loader produces same results for same input in different formats."""
    trajs = [
        types.Trajectory(
            obs=np.array([0, 1]),
            acts=np.array([100]),
            infos=None,
            terminal=True,
        ),
        types.Trajectory(
            obs=np.array([4, 5, 6]),
            acts=np.array([102, 103]),
            infos=None,
            terminal=True,
        ),
        types.Trajectory(
            obs=np.array([10, 11, 12, 13]),
            acts=np.array([104, 105, 106]),
            infos=None,
            terminal=False,
        ),
    ]
    trans = types.Transitions(
        obs=np.array([0, 4, 5, 10, 11, 12]),
        acts=np.array([100, 102, 103, 104, 105, 106]),
        next_obs=np.array([1, 5, 6, 11, 12, 13]),
        dones=np.array([True, False, True, False, False, False]),
        infos=np.array([{}] * 6),
    )
    trans_mapping = [
        {
            "obs": np.array([0, 4]),
            "acts": np.array([100, 102]),
            "next_obs": np.array([1, 5]),
            "dones": np.array([True, False]),
            "infos": np.array([{}, {}]),
        },
        {
            "obs": np.array([5, 10]),
            "acts": np.array([103, 104]),
            "next_obs": np.array([6, 11]),
            "dones": np.array([True, False]),
            "infos": np.array([{}, {}]),
        },
        {
            "obs": np.array([11, 12]),
            "acts": np.array([105, 106]),
            "next_obs": np.array([12, 13]),
            "dones": np.array([False]),
            "infos": np.array([{}, {}]),
        },
    ]

    for data in [trajs, trans, trans_mapping]:
        data_loader = base.make_data_loader(
            data,
            batch_size=2,
            data_loader_kwargs=dict(shuffle=False, drop_last=False),
        )
        for batch, expected_batch in zip(data_loader, trans_mapping):
            assert batch.keys() == expected_batch.keys()
            for k in batch.keys():
                v = batch[k]  # type: ignore[typed-dict-error]
                if isinstance(v, th.Tensor):
                    v = v.numpy()
                assert np.all(v == expected_batch[k])



"""This ingredient provides a reward network."""

import logging
import typing
from typing import Any, Mapping, Optional, Type

import sacred
from stable_baselines3.common import vec_env

from imitation.rewards import reward_nets
from imitation.util import networks

reward_ingredient = sacred.Ingredient("reward")
logger = logging.getLogger(__name__)


@reward_ingredient.config
def config():
    # Custom reward network
    net_cls = None
    net_kwargs = {}
    normalize_output_layer = networks.RunningNorm
    add_std_alpha = None
    ensemble_size = None
    ensemble_member_config = {}
    locals()  # quieten flake8


@reward_ingredient.named_config
def normalize_input_disable():
    net_kwargs = {"normalize_input_layer": None}  # noqa: F841


@reward_ingredient.named_config
def normalize_input_running():
    net_kwargs = {"normalize_input_layer": networks.RunningNorm}  # noqa: F841


@reward_ingredient.named_config
def normalize_output_disable():
    normalize_output_layer = None  # noqa: F841


@reward_ingredient.named_config
def normalize_output_running():
    normalize_output_layer = networks.RunningNorm  # noqa: F841


@reward_ingredient.named_config
def normalize_output_ema():
    normalize_output_layer = networks.EMANorm  # noqa: F841


@reward_ingredient.named_config
def reward_ensemble():
    net_cls = reward_nets.RewardEnsemble
    add_std_alpha = 0
    ensemble_size = 5
    normalize_output_layer = None
    ensemble_member_config = {
        "net_cls": reward_nets.BasicRewardNet,
        "net_kwargs": {},
        "normalize_output_layer": networks.RunningNorm,
    }
    locals()


@reward_ingredient.config_hook
def config_hook(config, command_name, logger):
    """Sets default values for `net_cls` and `net_kwargs`."""
    del logger
    res = {}
    if config["reward"]["net_cls"] is None:
        default_net = reward_nets.BasicRewardNet
        if command_name == "airl":
            default_net = reward_nets.BasicShapedRewardNet
        res["net_cls"] = default_net

    if "normalize_input_layer" not in config["reward"]["net_kwargs"]:
        res["net_kwargs"] = {"normalize_input_layer": networks.RunningNorm}

    if "net_cls" in res and issubclass(res["net_cls"], reward_nets.RewardEnsemble):
        del res["net_kwargs"]["normalize_input_layer"]

    return res


def _make_reward_net(
    venv: vec_env.VecEnv,
    net_cls: Type[reward_nets.RewardNet],
    net_kwargs: Mapping[str, Any],
    normalize_output_layer: Optional[Type[networks.BaseNorm]],
):
    """Helper function for creating reward nets."""
    reward_net = net_cls(
        venv.observation_space,
        venv.action_space,
        **net_kwargs,
    )

    if normalize_output_layer is not None:
        reward_net = reward_nets.NormalizedRewardNet(
            reward_net,
            normalize_output_layer,
        )

    return reward_net


@reward_ingredient.capture
def make_reward_net(
    venv: vec_env.VecEnv,
    net_cls: Type[reward_nets.RewardNet],
    net_kwargs: Mapping[str, Any],
    normalize_output_layer: Optional[Type[networks.BaseNorm]],
    add_std_alpha: Optional[float],
    ensemble_size: Optional[int],
    ensemble_member_config: Optional[Mapping[str, Any]],
) -> reward_nets.RewardNet:
    """Builds a reward network.

    Args:
        venv: Vectorized environment reward network will predict reward for.
        net_cls: Class of reward network to construct.
        net_kwargs: Keyword arguments passed to reward network constructor.
        normalize_output_layer: Wrapping the reward_net with NormalizedRewardNet
            to normalize the reward output.
        add_std_alpha: multiple of reward function standard deviation to add to the
            reward in predict_processed. Must be None when using a reward function that
            does not keep track of variance. Defaults to None.
        ensemble_size: The number of ensemble members to create. Must set if using
            `net_cls =` :class: `reward_nets.RewardEnsemble`.
        ensemble_member_config: The configuration for individual ensemble
            members. Note that `ensemble_member_config.net_cls` must not be
            :class: `reward_nets.RewardEnsemble`. Must be set if using
            `net_cls = ` :class: `reward_nets.RewardEnsemble`.

    Returns:
        A, possibly wrapped, instance of `net_cls`.

    Raises:
        ValueError: Using a reward ensemble but failed to provide configuration.
    """
    if issubclass(net_cls, reward_nets.RewardEnsemble):
        net_cls = typing.cast(Type[reward_nets.RewardEnsemble], net_cls)
        if ensemble_member_config is None:
            raise ValueError("Must specify ensemble_member_config.")

        if ensemble_size is None:
            raise ValueError("Must specify ensemble_size.")

        members = [
            _make_reward_net(venv, **ensemble_member_config)
            for _ in range(ensemble_size)
        ]

        reward_net: reward_nets.RewardNet = net_cls(
            venv.observation_space,
            venv.action_space,
            members,
        )

        if add_std_alpha is not None:
            assert isinstance(reward_net, reward_nets.RewardNetWithVariance)
            reward_net = reward_nets.AddSTDRewardWrapper(
                reward_net,
                default_alpha=add_std_alpha,
            )

        if normalize_output_layer is not None:
            raise ValueError("Output normalization not supported on RewardEnsembles.")

        return reward_net
    else:
        return _make_reward_net(venv, net_cls, net_kwargs, normalize_output_layer)



"""Utility functions for testing reward nets."""

import gymnasium as gym
import torch as th

from imitation.rewards import reward_nets


def make_ensemble(
    obs_space: gym.Space,
    action_space: gym.Space,
    num_members: int = 2,
    **kwargs,
):
    """Create a simple reward ensemble."""
    return reward_nets.RewardEnsemble(
        obs_space,
        action_space,
        members=[
            reward_nets.BasicRewardNet(obs_space, action_space, **kwargs)
            for _ in range(num_members)
        ],
    )


class MockRewardNet(reward_nets.RewardNet):
    """A mock reward net for testing."""

    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.Space,
        value: float = 0.0,
    ):
        """Create mock reward.

        Args:
            observation_space: observation space of the env
            action_space: action space of the env
            value: The reward to always return. Defaults to 0.0.
        """
        super().__init__(observation_space, action_space)
        self.value = value

    def forward(
        self,
        state: th.Tensor,
        action: th.Tensor,
        next_state: th.Tensor,
        done: th.Tensor,
    ) -> th.Tensor:
        batch_size = state.shape[0]
        return th.full(
            (batch_size,),
            fill_value=self.value,
            dtype=th.float32,
            device=state.device,
        )



"""Behavioural Cloning (BC).

Trains policy by applying supervised learning to a fixed dataset of (observation,
action) pairs generated by some expert demonstrator.
"""

import dataclasses
import itertools
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
)

import gymnasium as gym
import numpy as np
import torch as th
import tqdm
from stable_baselines3.common import policies, torch_layers, utils, vec_env

from imitation.algorithms import base as algo_base
from imitation.data import rollout, types
from imitation.policies import base as policy_base
from imitation.util import logger as imit_logger
from imitation.util import util


@dataclasses.dataclass(frozen=True)
class BatchIteratorWithEpochEndCallback:
    """Loops through batches from a batch loader and calls a callback after every epoch.

    Will throw an exception when an epoch contains no batches.
    """

    batch_loader: Iterable[types.TransitionMapping]
    n_epochs: Optional[int]
    n_batches: Optional[int]
    on_epoch_end: Optional[Callable[[int], None]]

    def __post_init__(self) -> None:
        epochs_and_batches_specified = (
            self.n_epochs is not None and self.n_batches is not None
        )
        neither_epochs_nor_batches_specified = (
            self.n_epochs is None and self.n_batches is None
        )
        if epochs_and_batches_specified or neither_epochs_nor_batches_specified:
            raise ValueError(
                "Must provide exactly one of `n_epochs` and `n_batches` arguments.",
            )

    def __iter__(self) -> Iterator[types.TransitionMapping]:
        def batch_iterator() -> Iterator[types.TransitionMapping]:
            # Note: the islice here ensures we do not exceed self.n_epochs
            for epoch_num in itertools.islice(itertools.count(), self.n_epochs):
                some_batch_was_yielded = False
                for batch in self.batch_loader:
                    yield batch
                    some_batch_was_yielded = True

                if not some_batch_was_yielded:
                    raise AssertionError(
                        f"Data loader returned no data during epoch "
                        f"{epoch_num} -- did it reset correctly?",
                    )
                if self.on_epoch_end is not None:
                    self.on_epoch_end(epoch_num)

        # Note: the islice here ensures we do not exceed self.n_batches
        return itertools.islice(batch_iterator(), self.n_batches)


@dataclasses.dataclass(frozen=True)
class BCTrainingMetrics:
    """Container for the different components of behavior cloning loss."""

    neglogp: th.Tensor
    entropy: Optional[th.Tensor]
    ent_loss: th.Tensor  # set to 0 if entropy is None
    prob_true_act: th.Tensor
    l2_norm: th.Tensor
    l2_loss: th.Tensor
    loss: th.Tensor


@dataclasses.dataclass(frozen=True)
class BehaviorCloningLossCalculator:
    """Functor to compute the loss used in Behavior Cloning."""

    ent_weight: float
    l2_weight: float

    def __call__(
        self,
        policy: policies.ActorCriticPolicy,
        obs: Union[
            types.AnyTensor,
            types.DictObs,
            Dict[str, np.ndarray],
            Dict[str, th.Tensor],
        ],
        acts: Union[th.Tensor, np.ndarray],
    ) -> BCTrainingMetrics:
        """Calculate the supervised learning loss used to train the behavioral clone.

        Args:
            policy: The actor-critic policy whose loss is being computed.
            obs: The observations seen by the expert.
            acts: The actions taken by the expert.

        Returns:
            A BCTrainingMetrics object with the loss and all the components it
            consists of.
        """
        tensor_obs = types.map_maybe_dict(
            util.safe_to_tensor,
            types.maybe_unwrap_dictobs(obs),
        )
        acts = util.safe_to_tensor(acts)

        # policy.evaluate_actions's type signatures are incorrect.
        # See https://github.com/DLR-RM/stable-baselines3/issues/1679
        (_, log_prob, entropy) = policy.evaluate_actions(
            tensor_obs,  # type: ignore[arg-type]
            acts,
        )
        prob_true_act = th.exp(log_prob).mean()
        log_prob = log_prob.mean()
        entropy = entropy.mean() if entropy is not None else None

        l2_norms = [th.sum(th.square(w)) for w in policy.parameters()]
        l2_norm = sum(l2_norms) / 2  # divide by 2 to cancel with gradient of square
        # sum of list defaults to float(0) if len == 0.
        assert isinstance(l2_norm, th.Tensor)

        ent_loss = -self.ent_weight * (entropy if entropy is not None else th.zeros(1))
        neglogp = -log_prob
        l2_loss = self.l2_weight * l2_norm
        loss = neglogp + ent_loss + l2_loss

        return BCTrainingMetrics(
            neglogp=neglogp,
            entropy=entropy,
            ent_loss=ent_loss,
            prob_true_act=prob_true_act,
            l2_norm=l2_norm,
            l2_loss=l2_loss,
            loss=loss,
        )


def enumerate_batches(
    batch_it: Iterable[types.TransitionMapping],
) -> Iterable[Tuple[Tuple[int, int, int], types.TransitionMapping]]:
    """Prepends batch stats before the batches of a batch iterator."""
    num_samples_so_far = 0
    for num_batches, batch in enumerate(batch_it):
        batch_size = len(batch["obs"])
        num_samples_so_far += batch_size
        yield (num_batches, batch_size, num_samples_so_far), batch


@dataclasses.dataclass(frozen=True)
class RolloutStatsComputer:
    """Computes statistics about rollouts.

    Args:
        venv: The vectorized environment in which to compute the rollouts.
        n_episodes: The number of episodes to base the statistics on.
    """

    venv: Optional[vec_env.VecEnv]
    n_episodes: int

    # TODO(shwang): Maybe instead use a callback that can be shared between
    #   all algorithms' `.train()` for generating rollout stats.
    #   EvalCallback could be a good fit:
    #   https://stable-baselines3.readthedocs.io/en/master/guide/callbacks.html#evalcallback

    def __call__(
        self,
        policy: policies.ActorCriticPolicy,
        rng: np.random.Generator,
    ) -> Mapping[str, float]:
        if self.venv is not None and self.n_episodes > 0:
            trajs = rollout.generate_trajectories(
                policy,
                self.venv,
                rollout.make_min_episodes(self.n_episodes),
                rng=rng,
            )
            return rollout.rollout_stats(trajs)
        else:
            return dict()


class BCLogger:
    """Utility class to help logging information relevant to Behavior Cloning."""

    def __init__(self, logger: imit_logger.HierarchicalLogger):
        """Create new BC logger.

        Args:
            logger: The logger to feed all the information to.
        """
        self._logger = logger
        self._tensorboard_step = 0
        self._current_epoch = 0

    def reset_tensorboard_steps(self):
        self._tensorboard_step = 0

    def log_epoch(self, epoch_number):
        self._current_epoch = epoch_number

    def log_batch(
        self,
        batch_num: int,
        batch_size: int,
        num_samples_so_far: int,
        training_metrics: BCTrainingMetrics,
        rollout_stats: Mapping[str, float],
    ):
        self._logger.record("batch_size", batch_size)
        self._logger.record("bc/epoch", self._current_epoch)
        self._logger.record("bc/batch", batch_num)
        self._logger.record("bc/samples_so_far", num_samples_so_far)
        for k, v in training_metrics.__dict__.items():
            self._logger.record(f"bc/{k}", float(v) if v is not None else None)

        for k, v in rollout_stats.items():
            if "return" in k and "monitor" not in k:
                self._logger.record("rollout/" + k, v)
        self._logger.dump(self._tensorboard_step)
        self._tensorboard_step += 1

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_logger"]
        return state


def reconstruct_policy(
    policy_path: str,
    device: Union[th.device, str] = "auto",
) -> policies.ActorCriticPolicy:
    """Reconstruct a saved policy.

    Args:
        policy_path: path where `.save_policy()` has been run.
        device: device on which to load the policy.

    Returns:
        policy: policy with reloaded weights.
    """
    policy = th.load(policy_path, map_location=utils.get_device(device))
    assert isinstance(policy, policies.ActorCriticPolicy)
    return policy


class BC(algo_base.DemonstrationAlgorithm):
    """Behavioral cloning (BC).

    Recovers a policy via supervised learning from observation-action pairs.
    """

    def __init__(
        self,
        *,
        observation_space: gym.Space,
        action_space: gym.Space,
        rng: np.random.Generator,
        policy: Optional[policies.ActorCriticPolicy] = None,
        demonstrations: Optional[algo_base.AnyTransitions] = None,
        batch_size: int = 32,
        minibatch_size: Optional[int] = None,
        optimizer_cls: Type[th.optim.Optimizer] = th.optim.Adam,
        optimizer_kwargs: Optional[Mapping[str, Any]] = None,
        ent_weight: float = 1e-3,
        l2_weight: float = 0.0,
        device: Union[str, th.device] = "auto",
        custom_logger: Optional[imit_logger.HierarchicalLogger] = None,
    ):
        """Builds BC.

        Args:
            observation_space: the observation space of the environment.
            action_space: the action space of the environment.
            rng: the random state to use for the random number generator.
            policy: a Stable Baselines3 policy; if unspecified,
                defaults to `FeedForward32Policy`.
            demonstrations: Demonstrations from an expert (optional). Transitions
                expressed directly as a `types.TransitionsMinimal` object, a sequence
                of trajectories, or an iterable of transition batches (mappings from
                keywords to arrays containing observations, etc).
            batch_size: The number of samples in each batch of expert data.
            minibatch_size: size of minibatch to calculate gradients over.
                The gradients are accumulated until `batch_size` examples
                are processed before making an optimization step. This
                is useful in GPU training to reduce memory usage, since
                fewer examples are loaded into memory at once,
                facilitating training with larger batch sizes, but is
                generally slower. Must be a factor of `batch_size`.
                Optional, defaults to `batch_size`.
            optimizer_cls: optimiser to use for supervised training.
            optimizer_kwargs: keyword arguments, excluding learning rate and
                weight decay, for optimiser construction.
            ent_weight: scaling applied to the policy's entropy regularization.
            l2_weight: scaling applied to the policy's L2 regularization.
            device: name/identity of device to place policy on.
            custom_logger: Where to log to; if None (default), creates a new logger.

        Raises:
            ValueError: If `weight_decay` is specified in `optimizer_kwargs` (use the
                parameter `l2_weight` instead), or if the batch size is not a multiple
                of the minibatch size.
        """
        self._demo_data_loader: Optional[Iterable[types.TransitionMapping]] = None
        self.batch_size = batch_size
        self.minibatch_size = minibatch_size or batch_size
        if self.batch_size % self.minibatch_size != 0:  # pragma: no cover
            raise ValueError("Batch size must be a multiple of minibatch size.")
        super().__init__(
            demonstrations=demonstrations,
            custom_logger=custom_logger,
        )
        self._bc_logger = BCLogger(self.logger)

        self.action_space = action_space
        self.observation_space = observation_space

        self.rng = rng

        if policy is None:
            extractor = (
                torch_layers.CombinedExtractor
                if isinstance(observation_space, gym.spaces.Dict)
                else torch_layers.FlattenExtractor
            )
            policy = policy_base.FeedForward32Policy(
                observation_space=observation_space,
                action_space=action_space,
                # Set lr_schedule to max value to force error if policy.optimizer
                # is used by mistake (should use self.optimizer instead).
                lr_schedule=lambda _: th.finfo(th.float32).max,
                features_extractor_class=extractor,
            )
        self._policy = policy.to(utils.get_device(device))
        # TODO(adam): make policy mandatory and delete observation/action space params?
        assert self.policy.observation_space == self.observation_space
        assert self.policy.action_space == self.action_space

        if optimizer_kwargs:
            if "weight_decay" in optimizer_kwargs:  # pragma: no cover
                raise ValueError("Use the parameter l2_weight instead of weight_decay.")
        optimizer_kwargs = optimizer_kwargs or {}
        self.optimizer = optimizer_cls(
            self.policy.parameters(),
            **optimizer_kwargs,
        )

        self.loss_calculator = BehaviorCloningLossCalculator(ent_weight, l2_weight)

    @property
    def policy(self) -> policies.ActorCriticPolicy:
        return self._policy

    def set_demonstrations(self, demonstrations: algo_base.AnyTransitions) -> None:
        self._demo_data_loader = algo_base.make_data_loader(
            demonstrations,
            self.minibatch_size,
        )

    def train(
        self,
        *,
        n_epochs: Optional[int] = None,
        n_batches: Optional[int] = None,
        on_epoch_end: Optional[Callable[[], None]] = None,
        on_batch_end: Optional[Callable[[], None]] = None,
        log_interval: int = 500,
        log_rollouts_venv: Optional[vec_env.VecEnv] = None,
        log_rollouts_n_episodes: int = 5,
        progress_bar: bool = True,
        reset_tensorboard: bool = False,
    ):
        """Train with supervised learning for some number of epochs.

        Here an 'epoch' is just a complete pass through the expert data loader,
        as set by `self.set_expert_data_loader()`. Note, that when you specify
        `n_batches` smaller than the number of batches in an epoch, the `on_epoch_end`
        callback will never be called.

        Args:
            n_epochs: Number of complete passes made through expert data before ending
                training. Provide exactly one of `n_epochs` and `n_batches`.
            n_batches: Number of batches loaded from dataset before ending training.
                Provide exactly one of `n_epochs` and `n_batches`.
            on_epoch_end: Optional callback with no parameters to run at the end of each
                epoch.
            on_batch_end: Optional callback with no parameters to run at the end of each
                batch.
            log_interval: Log stats after every log_interval batches.
            log_rollouts_venv: If not None, then this VecEnv (whose observation and
                actions spaces must match `self.observation_space` and
                `self.action_space`) is used to generate rollout stats, including
                average return and average episode length. If None, then no rollouts
                are generated.
            log_rollouts_n_episodes: Number of rollouts to generate when calculating
                rollout stats. Non-positive number disables rollouts.
            progress_bar: If True, then show a progress bar during training.
            reset_tensorboard: If True, then start plotting to Tensorboard from x=0
                even if `.train()` logged to Tensorboard previously. Has no practical
                effect if `.train()` is being called for the first time.
        """
        if reset_tensorboard:
            self._bc_logger.reset_tensorboard_steps()
        self._bc_logger.log_epoch(0)

        compute_rollout_stats = RolloutStatsComputer(
            log_rollouts_venv,
            log_rollouts_n_episodes,
        )

        def _on_epoch_end(epoch_number: int):
            if tqdm_progress_bar is not None:
                total_num_epochs_str = f"of {n_epochs}" if n_epochs is not None else ""
                tqdm_progress_bar.display(
                    f"Epoch {epoch_number} {total_num_epochs_str}",
                    pos=1,
                )
            self._bc_logger.log_epoch(epoch_number + 1)
            if on_epoch_end is not None:
                on_epoch_end()

        mini_per_batch = self.batch_size // self.minibatch_size
        n_minibatches = n_batches * mini_per_batch if n_batches is not None else None

        assert self._demo_data_loader is not None
        demonstration_batches = BatchIteratorWithEpochEndCallback(
            self._demo_data_loader,
            n_epochs,
            n_minibatches,
            _on_epoch_end,
        )
        batches_with_stats = enumerate_batches(demonstration_batches)
        tqdm_progress_bar: Optional[tqdm.tqdm] = None

        if progress_bar:
            batches_with_stats = tqdm.tqdm(
                batches_with_stats,
                unit="batch",
                total=n_minibatches,
            )
            tqdm_progress_bar = batches_with_stats

        def process_batch():
            self.optimizer.step()
            self.optimizer.zero_grad()

            if batch_num % log_interval == 0:
                rollout_stats = compute_rollout_stats(self.policy, self.rng)

                self._bc_logger.log_batch(
                    batch_num,
                    minibatch_size,
                    num_samples_so_far,
                    training_metrics,
                    rollout_stats,
                )

            if on_batch_end is not None:
                on_batch_end()

        self.optimizer.zero_grad()
        for (
            batch_num,
            minibatch_size,
            num_samples_so_far,
        ), batch in batches_with_stats:
            obs_tensor: Union[th.Tensor, Dict[str, th.Tensor]]
            # unwraps the observation if it's a dictobs and converts arrays to tensors
            obs_tensor = types.map_maybe_dict(
                lambda x: util.safe_to_tensor(x, device=self.policy.device),
                types.maybe_unwrap_dictobs(batch["obs"]),
            )
            acts = util.safe_to_tensor(batch["acts"], device=self.policy.device)
            training_metrics = self.loss_calculator(self.policy, obs_tensor, acts)

            # Renormalise the loss to be averaged over the whole
            # batch size instead of the minibatch size.
            # If there is an incomplete batch, its gradients will be
            # smaller, which may be helpful for stability.
            loss = training_metrics.loss * minibatch_size / self.batch_size
            loss.backward()

            batch_num = batch_num * self.minibatch_size // self.batch_size
            if num_samples_so_far % self.batch_size == 0:
                process_batch()
        if num_samples_so_far % self.batch_size != 0:
            # if there remains an incomplete batch
            batch_num += 1
            process_batch()



"""Tests `imitation.util.logger`."""

import csv
import json
import os.path as osp
from collections import defaultdict

import pytest
from stable_baselines3.common import logger as sb_logger

import imitation.util.logger as logger


def _csv_to_dict(csv_path: str) -> dict:
    result = defaultdict(list)
    with open(csv_path, "r") as f:
        for row in csv.DictReader(f):
            for k, v in row.items():
                if v != "":
                    v = float(v)
                result[k].append(v)
    return result


def _json_to_dict(json_path: str) -> dict:
    r"""Loads the saved json logging file and convert it to expected dict format.

    Args:
        json_path: Path of the json log file.
            Stored in the format - '{"A": 1, "B": 1}\n{"A": 2}\n{"B": 3}\n'

    Returns:
        dictionary in the format - `{"A": [1, 2, ""], "B": [1, "", 3]}`
    """
    result = defaultdict(list)
    with open(json_path, "r") as f:
        all_line_dicts = [json.loads(line) for line in f.readlines()]
    # get all the keys in the dict so as to add "" if the key is not present in a line
    all_keys = set().union(*[list(line_dict.keys()) for line_dict in all_line_dicts])

    for line_dict in all_line_dicts:
        for key in all_keys:
            result[key].append(line_dict.get(key, ""))
    return result


def _compare_csv_lines(csv_path: str, expect: dict):
    observed = _csv_to_dict(csv_path)
    assert expect == observed


def _compare_json_lines(json_path: str, expect: dict):
    observed = _json_to_dict(json_path)
    assert expect == observed


def test_no_accum(tmpdir):
    hier_logger = logger.configure(tmpdir, ["csv", "json"])
    assert hier_logger.get_dir() == tmpdir

    # Check that the recorded "A": -1 is overwritten by "A": 1 in the next line.
    # Previously, the observed value would be the mean of these two values (0) instead.
    hier_logger.record("A", -1)
    hier_logger.record("A", 1)
    hier_logger.record("B", 1)
    hier_logger.dump()

    hier_logger.record("A", 2)
    hier_logger.dump()
    hier_logger.record("B", 3)
    hier_logger.dump()
    expect = {"A": [1, 2, ""], "B": [1, "", 3]}
    _compare_csv_lines(osp.join(tmpdir, "progress.csv"), expect)
    _compare_json_lines(osp.join(tmpdir, "progress.json"), expect)


def test_raise_unknown_format():
    with pytest.raises(ValueError, match=r"Unknown format specified:.*"):
        logger.make_output_format("txt", "log_dir")


def test_free_form(tmpdir):
    hier_logger = logger.configure(tmpdir, ["log"])

    hier_logger.log("info 1")
    hier_logger.info("info 2")
    hier_logger.warn("warn 1")
    hier_logger.error("error 1")
    hier_logger.debug("debug not printed")
    hier_logger.set_level(level=sb_logger.DEBUG)
    hier_logger.debug("debug 1")
    with hier_logger.accumulate_means("foo"):
        hier_logger.info("info inner")
    hier_logger.debug("debug outer")

    hier_logger.close()

    with open(osp.join(tmpdir, "log.txt"), "r") as f:
        contents = f.readlines()
    assert contents == [
        "info 1\n",
        "info 2\n",
        "warn 1\n",
        "error 1\n",
        "debug 1\n",
        "info inner\n",
        "debug outer\n",
    ]


def test_reentry_fails(tmpdir):
    hier_logger = logger.configure(tmpdir)

    with hier_logger.accumulate_means("foo"):
        with pytest.raises(RuntimeError, match=r"Nested.*"):
            with hier_logger.accumulate_means("bar"):  # pragma: no cover
                pass


def test_close(tmpdir):
    hier_logger = logger.configure(tmpdir)

    hier_logger.record("A", 1)
    with hier_logger.accumulate_means("foo"):
        hier_logger.record("B", 2)
    hier_logger.dump()

    hier_logger.close()
    hier_logger.record("foo", 42)
    with pytest.raises(ValueError, match="I/O operation on closed file."):
        hier_logger.dump()


def test_name_to_value(tmpdir):
    hier_logger = logger.configure(tmpdir)

    with hier_logger.accumulate_means("foo"):
        hier_logger.record("A", 1)
        assert hier_logger.name_to_value["raw/foo/A"] == 1
        assert hier_logger.name_to_count["raw/foo/A"] == 0
        assert hier_logger.name_to_excluded["raw/foo/A"] is hier_logger.to_tuple(None)
        hier_logger.record("B", 10)
        assert hier_logger.name_to_value["raw/foo/B"] == 10
        assert hier_logger.name_to_count["raw/foo/B"] == 0
        assert hier_logger.name_to_excluded["raw/foo/B"] is hier_logger.to_tuple(None)
        hier_logger.dump()
        hier_logger.record("B", 20)
        assert hier_logger.name_to_value["raw/foo/B"] == 20
        assert hier_logger.name_to_count["raw/foo/B"] == 0
        assert hier_logger.name_to_excluded["raw/foo/B"] is hier_logger.to_tuple(None)
    assert hier_logger.name_to_value["mean/foo/A"] == 1
    assert hier_logger.name_to_count["mean/foo/A"] == 1
    assert hier_logger.name_to_excluded["mean/foo/A"] is hier_logger.to_tuple(None)
    assert hier_logger.name_to_value["mean/foo/B"] == 15
    assert hier_logger.name_to_count["mean/foo/B"] == 2
    assert hier_logger.name_to_excluded["mean/foo/B"] is hier_logger.to_tuple(None)
    hier_logger.dump()
    assert len(hier_logger.name_to_value) == 0
    assert len(hier_logger.name_to_count) == 0
    assert len(hier_logger.name_to_excluded) == 0


def test_hard(tmpdir):
    hier_logger = logger.configure(tmpdir)

    # Part One: Test logging outside the accumulating scope, and within scopes
    # with two different logging keys (including a repeat).

    hier_logger.record("no_context", 1)

    with hier_logger.accumulate_means("disc"):
        hier_logger.record("C", 2)
        hier_logger.record("D", 2)
        hier_logger.dump()
        hier_logger.record("C", 4)
        hier_logger.dump()

    with hier_logger.accumulate_means("gen"):
        hier_logger.record("E", 2)
        hier_logger.dump()
        hier_logger.record("E", 0)
        hier_logger.dump()

    with hier_logger.accumulate_means("disc"):
        hier_logger.record("C", 3)
        hier_logger.dump()

    hier_logger.dump()  # Writes 1 mean each from "gen" and "disc".

    expect_raw_gen = {"raw/gen/E": [2, 0]}
    expect_raw_disc = {
        "raw/disc/C": [2, 4, 3],
        "raw/disc/D": [2, "", ""],
    }
    expect_default = {
        "mean/gen/E": [1],
        "mean/disc/C": [3],
        "mean/disc/D": [2],
        "no_context": [1],
    }

    _compare_csv_lines(osp.join(tmpdir, "progress.csv"), expect_default)
    _compare_csv_lines(osp.join(tmpdir, "raw", "gen", "progress.csv"), expect_raw_gen)
    _compare_csv_lines(osp.join(tmpdir, "raw", "disc", "progress.csv"), expect_raw_disc)

    # Part Two:
    # Check that we append to the same logs after the first dump to "means/*".

    with hier_logger.accumulate_means("disc"):
        hier_logger.record("D", 100)
        hier_logger.dump()

    hier_logger.record("no_context", 2)

    hier_logger.dump()  # Writes 1 mean from "disc". "gen" is blank.

    expect_raw_gen = {"raw/gen/E": [2, 0]}
    expect_raw_disc = {
        "raw/disc/C": [2, 4, 3, ""],
        "raw/disc/D": [2, "", "", 100],
    }
    expect_default = {
        "mean/gen/E": [1, ""],
        "mean/disc/C": [3, ""],
        "mean/disc/D": [2, 100],
        "no_context": [1, 2],
    }

    _compare_csv_lines(osp.join(tmpdir, "progress.csv"), expect_default)
    _compare_csv_lines(osp.join(tmpdir, "raw", "gen", "progress.csv"), expect_raw_gen)
    _compare_csv_lines(osp.join(tmpdir, "raw", "disc", "progress.csv"), expect_raw_disc)


def test_accumulate_prefix(tmpdir):
    hier_logger = logger.configure(tmpdir)

    with hier_logger.add_accumulate_prefix("foo"), hier_logger.accumulate_means("bar"):
        hier_logger.record("A", 1)
        hier_logger.record("B", 2)
        hier_logger.dump()

    hier_logger.record("no_context", 1)

    with hier_logger.accumulate_means("blat"):
        hier_logger.record("C", 3)
        hier_logger.dump()

    hier_logger.dump()

    expect_raw_foo_bar = {
        "raw/foo/bar/A": [1],
        "raw/foo/bar/B": [2],
    }
    expect_raw_blat = {
        "raw/blat/C": [3],
    }
    expect_default = {
        "mean/foo/bar/A": [1],
        "mean/foo/bar/B": [2],
        "mean/blat/C": [3],
        "no_context": [1],
    }

    _compare_csv_lines(osp.join(tmpdir, "progress.csv"), expect_default)
    _compare_csv_lines(
        osp.join(tmpdir, "raw", "foo", "bar", "progress.csv"),
        expect_raw_foo_bar,
    )
    _compare_csv_lines(osp.join(tmpdir, "raw", "blat", "progress.csv"), expect_raw_blat)


def test_key_prefix(tmpdir):
    hier_logger = logger.configure(tmpdir)

    with hier_logger.accumulate_means("foo"), hier_logger.add_key_prefix("bar"):
        hier_logger.record("A", 1)
        hier_logger.record("B", 2)
        hier_logger.dump()

    hier_logger.record("no_context", 1)

    with hier_logger.accumulate_means("blat"):
        hier_logger.record("C", 3)
        hier_logger.dump()

    hier_logger.dump()

    expect_raw_foo_bar = {
        "raw/foo/bar/A": [1],
        "raw/foo/bar/B": [2],
    }
    expect_raw_blat = {
        "raw/blat/C": [3],
    }
    expect_default = {
        "mean/foo/bar/A": [1],
        "mean/foo/bar/B": [2],
        "mean/blat/C": [3],
        "no_context": [1],
    }

    _compare_csv_lines(osp.join(tmpdir, "progress.csv"), expect_default)
    _compare_csv_lines(
        osp.join(tmpdir, "raw", "foo", "progress.csv"),
        expect_raw_foo_bar,
    )
    _compare_csv_lines(osp.join(tmpdir, "raw", "blat", "progress.csv"), expect_raw_blat)


def test_cant_add_prefix_within_accumulate_means(tmpdir):
    h = logger.configure(tmpdir)
    with pytest.raises(RuntimeError):
        with h.accumulate_means("foo"), h.add_accumulate_prefix("bar"):
            pass  # pragma: no cover


def test_cant_add_key_prefix_outside_accumulate_means(tmpdir):
    h = logger.configure(tmpdir)
    with pytest.raises(RuntimeError):
        with h.add_key_prefix("bar"):
            pass  # pragma: no cover



"""This ingredient provides a number of logging utilities.

It is responsible for logging to WandB, TensorBoard, and stdout.
It will also create a symlink to the sacred logging directory in the log directory.
"""

import logging
import pathlib
from typing import Sequence, Tuple, Union

import huggingface_sb3 as hfsb3
import sacred

from imitation.scripts.ingredients import environment, wb
from imitation.util import logger as imit_logger
from imitation.util import sacred as sacred_util
from imitation.util import util

logging_ingredient = sacred.Ingredient(
    "logging",
    ingredients=[
        wb.wandb_ingredient,
        environment.environment_ingredient,
    ],
)
logger = logging.getLogger(__name__)


@logging_ingredient.config
def config():
    # Logging
    log_root = None
    log_dir = None
    log_level = logging.INFO
    log_format_strs = ["tensorboard", "stdout"]
    # The keys of log_format_strs_additional are concatenated to log_format_strs.
    # This allows named configs to add format strings, without changing the defaults.
    log_format_strs_additional = {}

    locals()  # silence flake8 unused variable warning


@logging_ingredient.config
def update_log_format_strs(log_format_strs, log_format_strs_additional):
    log_format_strs = log_format_strs + list(log_format_strs_additional.keys())


@logging_ingredient.config_hook
def hook(config, command_name: str, logger):
    del logger
    updates = {}
    if config["logging"]["log_dir"] is None:
        config_log_root = config["logging"]["log_root"] or "output"
        log_root = util.parse_path(config_log_root)
        env_sanitized = hfsb3.EnvironmentName(config["environment"]["gym_id"])
        assert isinstance(env_sanitized, str)
        log_dir = log_root / command_name / env_sanitized / util.make_unique_timestamp()
        updates["log_dir"] = log_dir
    return updates


@logging_ingredient.named_config
def wandb_logging():
    log_format_strs_additional = {"wandb": None}  # noqa: F841


@logging_ingredient.capture
def make_log_dir(
    _run,
    log_dir: str,
    log_level: Union[int, str],
) -> pathlib.Path:
    """Creates log directory and sets up symlink to Sacred logs.

    Args:
        log_dir: The directory to log to.
        log_level: The threshold of the logger. Either an integer level (10, 20, ...),
            a string of digits ('10', '20'), or a string of the designated level
            ('DEBUG', 'INFO', ...).

    Returns:
        The `log_dir`. This avoids the caller needing to capture this argument.
    """
    parsed_log_dir = util.parse_path(log_dir)
    parsed_log_dir.mkdir(parents=True, exist_ok=True)
    # convert strings of digits to numbers; but leave levels like 'INFO' unmodified
    try:
        log_level = int(log_level)
    except ValueError:
        pass
    logging.basicConfig(level=log_level)
    logger.info("Logging to %s", parsed_log_dir)
    sacred_util.build_sacred_symlink(parsed_log_dir, _run)
    return parsed_log_dir


@logging_ingredient.capture
def setup_logging(
    _run,
    log_format_strs: Sequence[str],
) -> Tuple[imit_logger.HierarchicalLogger, pathlib.Path]:
    """Builds the imitation logger.

    Args:
        log_format_strs: The types of formats to log to.

    Returns:
        The configured imitation logger and `log_dir`.
        Returning `log_dir` avoids the caller needing to capture this value.
    """
    log_dir = make_log_dir()
    if "wandb" in log_format_strs:
        wb.wandb_init(log_dir=str(log_dir))
    custom_logger = imit_logger.configure(
        folder=log_dir / "log",
        format_strs=log_format_strs,
    )
    return custom_logger, log_dir



"""Sphinx documentation configuration."""

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------
import io
import os
import urllib.request
import zipfile
from importlib import metadata

project = "imitation"
copyright = "2019-2022, Center for Human-Compatible AI"  # noqa: A001
author = "Center for Human-Compatible AI"

# The full version, including alpha/beta/rc tags
version = metadata.version("imitation")


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_github_changelog",
    "sphinx.ext.doctest",
    "myst_nb",
]

napoleon_google_docstring = True
napoleon_numpy_docstring = False
autosummary_generate = True

nb_execution_mode = os.getenv("NB_EXECUTION_MODE", "cache")
nb_execution_timeout = 200
nb_merge_streams = True
nb_output_stderr = "remove"
nb_execution_raise_on_error = True
nb_execution_show_tb = True

# The default engine ran into memory issues on some notebooks
# so we use lualatex instead
latex_engine = "lualatex"

# Enable LaTeX macros in markdown cells
myst_enable_extensions = [
    "amsmath",
    "dollarmath",
]

# skip_doctests is checked in our :skipif: directives in doctest examples
doctest_global_setup = """
import os

skip_doctests = os.getenv("SKIP_DOCTEST")
"""

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "special-members": "__init__",
    "show-inheritance": True,
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_title = "imitation"
html_theme_options = {
    "source_repository": "https://github.com/HumanCompatibleAI/imitation",
    "source_branch": "master",
    "source_directory": "docs",
    "light_css_variables": {
        "sidebar-item-font-size": "85%",
    },
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = ["css/custom.css"]

# -- Customization -----------------------------------------------------------


def no_namedtuple_attrib_docstring(app, what, name, obj, options, lines):
    """Remove redundant documentation in named tuples.

    Worksaround https://github.com/sphinx-doc/sphinx/issues/7353 -- adapted from
    https://chrisdown.name/2015/09/20/removing-namedtuple-docstrings-from-sphinx.html
    """  # noqa: DAR101
    is_namedtuple_docstring = 1 <= len(lines) <= 2 and lines[0].startswith(
        "Alias for field number",
    )
    if is_namedtuple_docstring:
        # We don't return, so we need to purge in-place
        del lines[:]


def setup(app):
    app.connect(
        "autodoc-process-docstring",
        no_namedtuple_attrib_docstring,
    )


# -- Download the latest benchmark summary -------------------------------------
download_url = (
    "https://github.com/HumanCompatibleAI/imitation/releases/latest/"
    "download/benchmark_runs.zip"
)

# Download the benchmark data, extract the summary and place it in the documentation
with urllib.request.urlopen(download_url) as url:
    with zipfile.ZipFile(io.BytesIO(url.read())) as z:
        with z.open("benchmark_runs/summary.md") as f:
            with open("main-concepts/benchmark_summary.md", "wb") as out:
                out.write(f.read())



"""Type alias shared by reward-related code."""

import abc
from typing import Protocol

import numpy as np


class RewardFn(Protocol):
    """Abstract class for reward function.

    Requires implementation of __call__() to compute the reward given a batch of
    states, actions, next states and dones.
    """

    @abc.abstractmethod
    def __call__(
        self,
        state: np.ndarray,
        action: np.ndarray,
        next_state: np.ndarray,
        done: np.ndarray,
    ) -> np.ndarray:
        """Compute rewards for a batch of transitions.

        Args:
            state: Current states of shape `(batch_size,) + state_shape`.
            action: Actions of shape `(batch_size,) + action_shape`.
            next_state: Successor states of shape `(batch_size,) + state_shape`.
            done: End-of-episode (terminal state) indicator of shape `(batch_size,)`.

        Returns:
            Computed rewards of shape `(batch_size,`).
        """  # noqa: DAR202



#!/usr/bin/env python
"""Clean all notebooks in the repository."""
import argparse
import pathlib
import sys
import traceback
from typing import Any, Dict, List

import nbformat


class UncleanNotebookError(Exception):
    """Raised when a notebook is unclean."""


markdown_structure: Dict[str, Dict[str, Any]] = {
    "cell_type": {"do": "keep"},
    "metadata": {"do": "constant", "value": dict()},
    "source": {"do": "keep"},
    "id": {"do": "keep"},
}

code_structure: Dict[str, Dict[str, Any]] = {
    "cell_type": {"do": "keep"},
    "metadata": {"do": "constant", "value": dict()},
    "source": {"do": "keep"},
    "outputs": {"do": "constant", "value": list()},
    "execution_count": {"do": "constant", "value": None},
    "id": {"do": "keep"},
}

structure: Dict[str, Dict[str, Dict[str, Any]]] = {
    "markdown": markdown_structure,
    "code": code_structure,
}


def clean_notebook(file: pathlib.Path, check_only=False) -> None:
    """Clean an ipynb notebook.

    "Cleaning" means removing all output and metadata, as well as any other unnecessary
    or vendor-dependent information or fields, so that it can be committed to the
    repository, and so that artificial diffs are not introduced when the notebook is
    executed.

    Args:
        file: Path to the notebook to clean.
        check_only: If True, only check if the notebook is clean, and raise an
            exception if it is not. If False, clean the notebook in-place.

    Raises:
        UncleanNotebookError: If `check_only` is True and the notebook is not clean.
            Message contains brief description of the reason for the failure.
        ValueError: unknown cell structure action.
    """
    # Read the notebook
    with open(file) as f:
        nb = nbformat.read(f, as_version=4)

    was_dirty = False

    if check_only:
        print(f"Checking {file}")

    for cell in nb.cells:
        # Remove empty cells
        if cell["cell_type"] == "code" and not cell["source"]:
            if check_only:
                raise UncleanNotebookError(f"Notebook {file} has empty code cell")
            nb.cells.remove(cell)
            was_dirty = True

        # Clean the cell
        # (copy the cell keys list so we can iterate over it while modifying it)
        for key in list(cell):
            if key not in structure[cell["cell_type"]]:
                if check_only:
                    raise UncleanNotebookError(
                        f"Notebook {file} has unknown cell key {key}",
                    )
                del cell[key]
                was_dirty = True
            else:
                cell_structure = structure[cell["cell_type"]][key]
                if cell_structure["do"] == "keep":
                    continue
                elif cell_structure["do"] == "constant":
                    constant_value = cell_structure["value"]
                    if cell[key] != constant_value:
                        if check_only:
                            raise UncleanNotebookError(
                                f"Notebook {file} has illegal cell value for key {key}"
                                f" (value: {cell[key]}, "
                                f"expected: {constant_value})",
                            )
                        cell[key] = constant_value
                        was_dirty = True
                else:
                    raise ValueError(
                        f"Unknown cell structure action {cell_structure['do']}",
                    )

    if not check_only and was_dirty:
        # Write the notebook
        with open(file, "w") as f:
            nbformat.write(nb, f)
        print(f"Cleaned {file}")


def parse_args():
    """Parse command-line arguments."""
    # if the argument --check has been passed, check if the notebooks are clean
    # otherwise, clean them in-place
    parser = argparse.ArgumentParser()
    # capture files and paths to clean
    parser.add_argument(
        "files",
        nargs="+",
        type=pathlib.Path,
        help="List of files or paths to clean",
    )
    parser.add_argument("--check", action="store_true")
    args = parser.parse_args()
    return parser, args


def get_files(input_paths: List):
    """Build list of files to scan from list of paths and files."""
    files = []
    for file in input_paths:
        if file.is_dir():
            files.extend(file.glob("**/*.ipynb"))
        else:
            if file.suffix == ".ipynb":
                files.append(file)
            else:
                print(f"Skipping {file} (not a notebook)")
    if not files:
        print("No notebooks found")
        sys.exit(1)
    return files


def main():
    """Clean all notebooks in the repository, or check that they are clean."""
    parser, args = parse_args()
    check_only = args.check
    input_paths = args.files

    if len(input_paths) == 0:
        parser.print_help()
        sys.exit(1)

    files = get_files(input_paths)

    for file in files:
        try:
            clean_notebook(file, check_only=check_only)
        except UncleanNotebookError:
            traceback.print_exc()
            sys.exit(1)


if __name__ == "__main__":
    main()



"""Helper methods to build and run neural networks."""
import abc
import collections
import contextlib
import functools
from typing import Dict, Iterable, Optional, Type, Union

import torch as th
from torch import nn


@contextlib.contextmanager
def training_mode(m: nn.Module, mode: bool = False):
    """Temporarily switch module ``m`` to specified training ``mode``.

    Args:
        m: The module to switch the mode of.
        mode: whether to set training mode (``True``) or evaluation (``False``).

    Yields:
        The module `m`.
    """
    # Modified from Christoph Heindl's method posted on:
    # https://discuss.pytorch.org/t/opinion-eval-should-be-a-context-manager/18998/3
    old_mode = m.training
    m.train(mode)
    try:
        yield m
    finally:
        m.train(old_mode)


training = functools.partial(training_mode, mode=True)
evaluating = functools.partial(training_mode, mode=False)


class SqueezeLayer(nn.Module):
    """Torch module that squeezes a B*1 tensor down into a size-B vector."""

    def forward(self, x):
        assert x.ndim == 2 and x.shape[1] == 1
        new_value = x.squeeze(1)
        assert new_value.ndim == 1
        return new_value


class BaseNorm(nn.Module, abc.ABC):
    """Base class for layers that try to normalize the input to mean 0 and variance 1.

    Similar to BatchNorm, LayerNorm, etc. but whereas they only use statistics from
    the current batch at train time, we use statistics from all batches.
    """

    running_mean: th.Tensor
    running_var: th.Tensor
    count: th.Tensor

    def __init__(self, num_features: int, eps: float = 1e-5):
        """Builds RunningNorm.

        Args:
            num_features: Number of features; the length of the non-batch dimension.
            eps: Small constant for numerical stability. Inputs are rescaled by
                `1 / sqrt(estimated_variance + eps)`.
        """
        super().__init__()
        self.eps = eps
        self.register_buffer("running_mean", th.empty(num_features))
        self.register_buffer("running_var", th.empty(num_features))
        self.register_buffer("count", th.empty((), dtype=th.int))
        BaseNorm.reset_running_stats(self)

    def reset_running_stats(self) -> None:
        """Resets running stats to defaults, yielding the identity transformation."""
        self.running_mean.zero_()
        self.running_var.fill_(1)
        self.count.zero_()

    def forward(self, x: th.Tensor) -> th.Tensor:
        """Updates statistics if in training mode. Returns normalized `x`."""
        if self.training:
            # Do not backpropagate through updating running mean and variance.
            # These updates are in-place and not differentiable. The gradient
            # is not needed as the running mean and variance are updated
            # directly by this function, and not by gradient descent.
            with th.no_grad():
                self.update_stats(x)

        # Note: this is different from the behavior in stable-baselines, see
        # https://github.com/HumanCompatibleAI/imitation/issues/442
        return (x - self.running_mean) / th.sqrt(self.running_var + self.eps)

    @abc.abstractmethod
    def update_stats(self, batch: th.Tensor) -> None:
        """Update `self.running_mean`, `self.running_var` and `self.count`."""


class RunningNorm(BaseNorm):
    """Normalizes input to mean 0 and standard deviation 1 using a running average.

    Similar to BatchNorm, LayerNorm, etc. but whereas they only use statistics from
    the current batch at train time, we use statistics from all batches.

    This should replicate the common practice in RL of normalizing environment
    observations, such as using ``VecNormalize`` in Stable Baselines. Note that
    the behavior of this class is slightly different from `VecNormalize`, e.g.,
    it works with the current reward instead of return estimate, and subtracts the mean
    reward whereas ``VecNormalize`` only rescales it.
    """

    def update_stats(self, batch: th.Tensor) -> None:
        """Update `self.running_mean`, `self.running_var` and `self.count`.

        Uses Chan et al (1979), "Updating Formulae and a Pairwise Algorithm for
        Computing Sample Variances." to update the running moments in a numerically
        stable fashion.

        Args:
            batch: A batch of data to use to update the running mean and variance.
        """
        batch_mean = th.mean(batch, dim=0)
        batch_var = th.var(batch, dim=0, unbiased=False)
        batch_count = batch.shape[0]

        delta = batch_mean - self.running_mean
        tot_count = self.count + batch_count
        self.running_mean += delta * batch_count / tot_count

        self.running_var *= self.count
        self.running_var += batch_var * batch_count
        self.running_var += th.square(delta) * self.count * batch_count / tot_count
        self.running_var /= tot_count

        self.count += batch_count


class EMANorm(BaseNorm):
    """Similar to RunningNorm but uses an exponential weighting."""

    inv_learning_rate: th.Tensor
    num_batches: th.IntTensor

    def __init__(
        self,
        num_features: int,
        decay: float = 0.99,
        eps: float = 1e-5,
    ):
        """Builds EMARunningNorm.

        Args:
            num_features: Number of features; the length of the non-batch dim.
            decay: how quickly the weight on past samples decays over time.
            eps: small constant for numerical stability.

        Raises:
            ValueError: if decay is out of range.
        """
        super().__init__(num_features, eps=eps)

        if not 0 < decay < 1:
            raise ValueError("decay must be between 0 and 1")

        self.decay = decay
        self.register_buffer("inv_learning_rate", th.empty(()))
        self.register_buffer("num_batches", th.empty((), dtype=th.int))
        EMANorm.reset_running_stats(self)

    def reset_running_stats(self):
        """Reset the running stats of the normalization layer."""
        super().reset_running_stats()
        self.inv_learning_rate.zero_()
        self.num_batches.zero_()

    def update_stats(self, batch: th.Tensor) -> None:
        """Update `self.running_mean` and `self.running_var` in batch mode.

        Reference Algorithm 3 from:
        https://github.com/HumanCompatibleAI/imitation/files/9456540/Incremental_batch_EMA_and_EMV.pdf

        Args:
            batch: A batch of data to use to update the running mean and variance.
        """
        b_size = batch.shape[0]
        if len(batch.shape) == 1:
            batch = batch.reshape(b_size, 1)

        self.inv_learning_rate += self.decay**self.num_batches
        learning_rate = 1 / self.inv_learning_rate

        # update running mean
        delta_mean = batch.mean(0) - self.running_mean
        self.running_mean += learning_rate * delta_mean

        # update running variance
        batch_var = batch.var(0, unbiased=False)
        delta_var = batch_var + (1 - learning_rate) * delta_mean**2 - self.running_var
        self.running_var += learning_rate * delta_var

        self.count += b_size
        self.num_batches += 1  # type: ignore[misc]


def build_mlp(
    in_size: int,
    hid_sizes: Iterable[int],
    out_size: int = 1,
    name: Optional[str] = None,
    activation: Type[nn.Module] = nn.ReLU,
    dropout_prob: float = 0.0,
    squeeze_output: bool = False,
    flatten_input: bool = False,
    normalize_input_layer: Optional[Type[nn.Module]] = None,
) -> nn.Module:
    """Constructs a Torch MLP.

    Args:
        in_size: size of individual input vectors; input to the MLP will be of
            shape (batch_size, in_size).
        hid_sizes: sizes of hidden layers. If this is an empty iterable, then we build
            a linear function approximator.
        out_size: size of output vector.
        name: Name to use as a prefix for the layers ID.
        activation: activation to apply after hidden layers.
        dropout_prob: Dropout probability to use after each hidden layer. If 0,
            no dropout layers are added to the network.
        squeeze_output: if out_size=1, then squeeze_input=True ensures that MLP
            output is of size (B,) instead of (B,1).
        flatten_input: should input be flattened along axes 1, 2, 3, …? Useful
            if you want to, e.g., process small images inputs with an MLP.
        normalize_input_layer: if specified, module to use to normalize inputs;
            e.g. `nn.BatchNorm` or `RunningNorm`.

    Returns:
        nn.Module: an MLP mapping from inputs of size (batch_size, in_size) to
            (batch_size, out_size), unless out_size=1 and squeeze_output=True,
            in which case the output is of size (batch_size, ).

    Raises:
        ValueError: if squeeze_output was supplied with out_size!=1.
    """
    layers: Dict[str, nn.Module] = {}

    if name is None:
        prefix = ""
    else:
        prefix = f"{name}_"

    if flatten_input:
        layers[f"{prefix}flatten"] = nn.Flatten()

    # Normalize input layer
    if normalize_input_layer:
        try:
            layer_instance = normalize_input_layer(in_size)  # type: ignore[call-arg]
        except TypeError as exc:
            raise ValueError(
                f"normalize_input_layer={normalize_input_layer} is not a valid "
                "normalization layer type accepting only one argument (in_size).",
            ) from exc
        layers[f"{prefix}normalize_input"] = layer_instance

    # Hidden layers
    prev_size = in_size
    for i, size in enumerate(hid_sizes):
        layers[f"{prefix}dense{i}"] = nn.Linear(prev_size, size)
        prev_size = size
        if activation:
            layers[f"{prefix}act{i}"] = activation()
        if dropout_prob > 0.0:
            layers[f"{prefix}dropout{i}"] = nn.Dropout(dropout_prob)

    # Final dense layer
    layers[f"{prefix}dense_final"] = nn.Linear(prev_size, out_size)

    if squeeze_output:
        if out_size != 1:
            raise ValueError("squeeze_output is only applicable when out_size=1")
        layers[f"{prefix}squeeze"] = SqueezeLayer()

    model = nn.Sequential(collections.OrderedDict(layers))

    return model


def build_cnn(
    in_channels: int,
    hid_channels: Iterable[int],
    out_size: int = 1,
    name: Optional[str] = None,
    activation: Type[nn.Module] = nn.ReLU,
    kernel_size: int = 3,
    stride: int = 1,
    padding: Union[int, str] = "same",
    dropout_prob: float = 0.0,
    squeeze_output: bool = False,
) -> nn.Module:
    """Constructs a Torch CNN.

    Args:
        in_channels: number of channels of individual inputs; input to the CNN will have
            shape (batch_size, in_size, in_height, in_width).
        hid_channels: number of channels of hidden layers. If this is an empty iterable,
            then we build a linear function approximator.
        out_size: size of output vector.
        name: Name to use as a prefix for the layers ID.
        activation: activation to apply after hidden layers.
        kernel_size: size of convolutional kernels.
        stride: stride of convolutional kernels.
        padding: padding of convolutional kernels.
        dropout_prob: Dropout probability to use after each hidden layer. If 0,
            no dropout layers are added to the network.
        squeeze_output: if out_size=1, then squeeze_input=True ensures that CNN
            output is of size (B,) instead of (B,1).

    Returns:
        nn.Module: a CNN mapping from inputs of size (batch_size, in_size, in_height,
            in_width) to (batch_size, out_size), unless out_size=1 and
            squeeze_output=True, in which case the output is of size (batch_size, ).

    Raises:
        ValueError: if squeeze_output was supplied with out_size!=1.
    """
    layers: Dict[str, nn.Module] = {}

    if name is None:
        prefix = ""
    else:
        prefix = f"{name}_"

    prev_channels = in_channels
    for i, n_channels in enumerate(hid_channels):
        layers[f"{prefix}conv{i}"] = nn.Conv2d(
            prev_channels,
            n_channels,
            kernel_size,
            stride=stride,
            padding=padding,
        )
        prev_channels = n_channels
        if activation:
            layers[f"{prefix}act{i}"] = activation()
        if dropout_prob > 0.0:
            layers[f"{prefix}dropout{i}"] = nn.Dropout(dropout_prob)

    # final dense layer
    layers[f"{prefix}avg_pool"] = nn.AdaptiveAvgPool2d(1)
    layers[f"{prefix}flatten"] = nn.Flatten()
    layers[f"{prefix}dense_final"] = nn.Linear(prev_channels, out_size)

    if squeeze_output:
        if out_size != 1:
            raise ValueError("squeeze_output is only applicable when out_size=1")
        layers[f"{prefix}squeeze"] = SqueezeLayer()

    model = nn.Sequential(collections.OrderedDict(layers))
    return model



"""Tests `imitation.util.sacred` and `imitation.util.util`."""

import warnings

import numpy as np
import pytest
import torch as th
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.extra.numpy import arrays

from imitation.util import sacred as sacred_util
from imitation.util import util


def test_endless_iter():
    x = range(2)
    it = util.endless_iter(x)
    assert next(it) == 0
    assert next(it) == 1
    assert next(it) == 0
    assert next(it) == 1
    assert next(it) == 0


def test_endless_iter_error():
    x = []
    with pytest.raises(ValueError, match="no elements"):
        util.endless_iter(x)
    with pytest.raises(ValueError, match="needs a non-iterator Iterable"):
        generator = (x for x in range(5))
        util.endless_iter(generator)


@given(
    st.lists(
        st.integers(),
        min_size=1,
    ),
)
def test_get_first_iter_element(input_seq):
    with pytest.raises(ValueError, match="iterable.* had no elements"):
        util.get_first_iter_element([])

    first_element, new_iterable = util.get_first_iter_element(input_seq)
    assert first_element == input_seq[0]
    assert input_seq is new_iterable

    def generator_fn():
        for x in input_seq:
            yield x

    generator = generator_fn()
    assert generator == iter(generator)
    first_element, new_iterable = util.get_first_iter_element(generator)
    assert first_element == input_seq[0]
    assert list(new_iterable) == input_seq
    assert list(new_iterable) == []


@given(
    arrays(
        dtype=np.float64,
        shape=st.integers(min_value=1, max_value=10),
        elements=st.floats(min_value=1e-3, max_value=1e6, allow_nan=False),
    ).map(
        # Compute the fractional part of the sum of the elements, divide it by
        # the number of elements, and subtract this from every element.
        # This ensures that the sum of the elements is integral.
        lambda x: (x - (x.sum() - np.floor(x.sum())) / len(x)),
    ),
)
def test_integer_constrained_rounding(x: np.ndarray):
    original_sum = x.sum()

    rounded = util.oric(x)
    assert np.allclose(rounded.sum(), original_sum)
    assert np.abs(x - rounded).max() <= 1.0


def test_dict_get_nested():
    assert sacred_util.dict_get_nested({}, "asdf.foo", default=4) == 4
    assert sacred_util.dict_get_nested({"a": {"b": "c"}}, "a.b") == "c"


def test_safe_to_tensor():
    # Copy iff the array is non-writable.
    numpy = np.array([1, 2, 3])
    torch = util.safe_to_tensor(numpy)
    assert np.may_share_memory(numpy, torch)

    # This should never cause a warning from `th.as_tensor`.
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        numpy.flags.writeable = False
        torch = util.safe_to_tensor(numpy)
        assert not np.may_share_memory(numpy, torch)


def test_safe_to_numpy():
    tensor = th.tensor([1, 2, 3])
    numpy = util.safe_to_numpy(tensor)
    assert (numpy == tensor.numpy()).all()
    assert util.safe_to_numpy(None) is None
    with pytest.warns(UserWarning, match=".*performance.*"):
        util.safe_to_numpy(tensor, warn=True)


def test_tensor_iter_norm():
    # vector is [1,0,1,1,-5,-6]; its 2-norm is 8, and 1-norm is 14
    tensor_list = [
        th.tensor([1.0, 0.0]),
        th.tensor([[1.0], [1.0], [-5.0]]),
        th.tensor([-6.0]),
    ]
    norm_2 = util.tensor_iter_norm(tensor_list, ord=2).item()
    assert np.allclose(norm_2, 8.0)
    norm_1 = util.tensor_iter_norm(tensor_list, ord=1).item()
    assert np.allclose(norm_1, 14.0)
    with pytest.raises(ValueError):
        util.tensor_iter_norm(tensor_list, ord=0.0)



"""Fixtures common across tests."""

import numpy as np
import pytest
import seals  # noqa: F401
import torch
from stable_baselines3.common.vec_env import VecEnv

from imitation.data.wrappers import RolloutInfoWrapper
from imitation.util import logger, util

CARTPOLE_ENV_NAME = "seals/CartPole-v0"


@pytest.fixture(params=[1, 4], ids=lambda n: f"vecenv({n})")
def cartpole_venv(request, rng) -> VecEnv:
    num_envs = request.param
    return util.make_vec_env(
        CARTPOLE_ENV_NAME,
        n_envs=num_envs,
        post_wrappers=[lambda env, _: RolloutInfoWrapper(env)],
        rng=rng,
    )


@pytest.fixture(scope="session", autouse=True)
def torch_single_threaded():
    """Make PyTorch execute code single-threaded.

    This allows us to run the test suite with greater across-test parallelism.
    This is faster, since:
        - There are diminishing returns to more threads within a test.
        - Many tests cannot be multi-threaded (e.g. most not using PyTorch training),
          and we have to set between-test parallelism based on peak resource
          consumption of tests to avoid spurious failures.
    """
    torch.set_num_threads(1)
    torch.set_num_interop_threads(1)


@pytest.fixture()
def custom_logger(tmpdir: str) -> logger.HierarchicalLogger:
    return logger.configure(tmpdir)


@pytest.fixture()
def rng() -> np.random.Generator:
    return np.random.default_rng(seed=0)



"""Tests `imitation.rewards.reward_function` and `imitation.rewards.serialize`."""

import numpy as np
import pytest

from imitation.rewards import reward_function, serialize

OBS = np.random.randint(0, 10, (64, 100))
ACTS = NEXT_OBS = OBS
DONES = np.zeros(64, dtype=np.bool_)


def _funky_reward_fn(obs, act, next_obs, done):
    """Returns consecutive reward from 1 to batch size `len(obs)`."""
    # give each environment number from 1 to num_envs
    return (np.arange(len(obs))).astype("float32")


def _invalid_reward_fn(obs, act, next_obs, done):
    """Returns rewards for lesser number of observations."""
    return (np.arange(len(obs) - 1)).astype("float32")


def test_reward_fn_override():
    # test inheriting class from RewardFn works
    class InheritedFunkyReward(reward_function.RewardFn):
        """A reward inherited from RewardFn."""

        def __init__(self):
            super().__init__()

        def __call__(self, obs, act, next_obs, steps=None):
            """Returns consecutive reward from 0 to batch size -1 (`len(obs)` - 1)."""
            return (np.arange(len(obs))).astype("float32")

    inherited_funky_reward_fn = InheritedFunkyReward()
    inherited_funky_reward_fn(OBS, ACTS, NEXT_OBS)


def test_validate_rewardfn_class():
    validated_reward_fn = serialize.ValidateRewardFn(_funky_reward_fn)
    validated_reward_fn(OBS, ACTS, NEXT_OBS, DONES)

    with pytest.raises(AssertionError):
        invalidated_reward_fn = serialize.ValidateRewardFn(_invalid_reward_fn)
        invalidated_reward_fn(OBS, ACTS, NEXT_OBS, DONES)



"""Tests for imitation.data.huggingface_utils."""
import datetime
import pathlib
from typing import Sequence, cast

import hypothesis
import hypothesis.strategies as st
import numpy as np

import imitation.testing.hypothesis_strategies as h_strats
from imitation.data import huggingface_utils, serialize, types


def wrap_in_trajectory_dataset_sequence(
    trajectories: Sequence[types.Trajectory],
) -> huggingface_utils.TrajectoryDatasetSequence:
    """Takes a list of trajectories in a TrajectoryDatasetSequence.

    It first converts the trajectories to a dataset and then presents that
    dataset as a sequence of trajectories again using TrajectoryDatasetSequence.

    This is useful for testing the TrajectoryDatasetSequence class without having to
    load a dataset from the hub or disk.

    Args:
        trajectories: The trajectories to wrap.

    Returns:
        The wrapped trajectories.
    """
    return huggingface_utils.TrajectoryDatasetSequence(
        huggingface_utils.trajectories_to_dataset(trajectories),
    )


@hypothesis.given(trajectories=h_strats.trajectories_list, do_wrap=st.booleans())
@hypothesis.settings(
    suppress_health_check=[
        # Note: needed to convince hypothesis that it is ok that we use the same
        #   temporary directory for all tests.
        hypothesis.HealthCheck.function_scoped_fixture,
    ],
    deadline=datetime.timedelta(seconds=0.5),  # reduce flakiness
)
def test_save_load_roundtrip(
    trajectories: Sequence[types.Trajectory],
    do_wrap: bool,
    tmpdir: pathlib.Path,
):
    """Test saving and loading a sequence of trajectories."""
    # GIVEN
    if do_wrap:
        trajectories = wrap_in_trajectory_dataset_sequence(trajectories)

    # WHEN
    serialize.save(tmpdir, trajectories)
    loaded_trajectories = serialize.load(tmpdir)

    # THEN
    assert len(trajectories) == len(loaded_trajectories)
    for traj, loaded_traj in zip(trajectories, loaded_trajectories):
        assert traj == loaded_traj


@hypothesis.given(st.data(), h_strats.trajectories_list)
# the first run sometimes takes longer, so we give it more time
@hypothesis.settings(deadline=datetime.timedelta(milliseconds=300))
def test_sliced_access(data: st.DataObject, trajectories: Sequence[types.Trajectory]):
    """Test that slicing a TrajectoryDatasetSequence behaves as expected."""
    # GIVEN
    wrapped_trajectories = wrap_in_trajectory_dataset_sequence(trajectories)
    slices_strategy = st.slices(len(trajectories))

    # Note: we test if for 10 slices at a time because creating the dataset is slow
    for _ in range(10):
        # GIVEN
        the_slice = data.draw(slices_strategy)
        indices_of_slice = list(range(*the_slice.indices(len(trajectories))))

        # WHEN
        sliced_data = wrapped_trajectories[the_slice]

        # THEN
        for idx, traj in zip(indices_of_slice, sliced_data):
            assert traj == trajectories[idx]


@hypothesis.given(st.data(), h_strats.trajectory)
# the first run sometimes takes longer, so we give it more time
@hypothesis.settings(deadline=datetime.timedelta(milliseconds=300))
def test_sliced_info_dict_access(
    data: st.DataObject,
    trajectory: types.Trajectory,
) -> None:
    """Test slicing of info dicts from a TrajectoryDatasetSequence.

    Note: this test ensures coverage of sliced access to info dicts,
    which are wrapped in a `_LazyDecodedList`.

    Args:
        data: The hypothesis data object.
        trajectory: The trajectory to test slicing on.
    """
    # GIVEN
    # Note: we need to add the cast here since for some reason, the type of the
    # __getitem__ method is not inferred correctly.
    wrapped_trajectory = cast(
        types.Trajectory,
        wrap_in_trajectory_dataset_sequence([trajectory])[0],
    )
    # Note: the cast of the info dicts is needed since the `infos` field of a
    #  trajectory is optional.
    info_dicts = cast(np.ndarray, trajectory.infos)
    wrapped_info_dicts = cast(np.ndarray, wrapped_trajectory.infos)
    slices_strategy = st.slices(len(wrapped_info_dicts))

    # Note: we test if for 10 slices at a time because creating the dataset is slow
    for _ in range(10):
        # GIVEN
        the_slice = data.draw(slices_strategy)
        indices_of_slice = list(range(*the_slice.indices(len(wrapped_info_dicts))))

        # WHEN
        info_slice = wrapped_info_dicts[the_slice]

        # THEN
        for idx, info_dict in zip(indices_of_slice, info_slice):
            assert info_dict == info_dicts[idx]



"""This ingredient provides an expert policy.

The expert policy is either loaded from disk or from the HuggingFace Model Hub or is
a test policy (e.g., random or zero).
The supported policy types are:

- :code:`ppo` and :code:`sac`: A policy trained with SB3.
    Needs a `path` in the `loader_kwargs`.
- :code:`<algo>-huggingface` (algo can be `ppo` or `sac`):
    A policy trained with SB3 and uploaded to the HuggingFace Model Hub.
    Will load the model from the repo :code:`<organization>/<algo>-<env_name>`.
    You can set the organization with the `organization` key in :code:`loader_kwargs`.
    The default is `HumanCompatibleAI`.
- :code:`random`: A policy that takes random actions.
- :code:`zero`: A policy that takes zero actions.
"""
import sacred

from imitation.policies import serialize
from imitation.scripts.ingredients import environment

expert_ingredient = sacred.Ingredient(
    "expert",
    ingredients=[environment.environment_ingredient],
)


@expert_ingredient.config
def config():
    # [ppo, sac, random, zero, ppo-huggingface, sac-huggingface] or your own.
    policy_type = "ppo-huggingface"
    # See imitation.policies.serialize.load_policy for options.
    loader_kwargs = dict()
    locals()  # quieten flake8


@expert_ingredient.config_hook
def config_hook(config, command_name, logger):
    e_config = config["expert"]
    if "huggingface" in e_config["policy_type"]:
        # Set the default loader_kwargs for huggingface policies.
        if "organization" not in e_config["loader_kwargs"]:
            e_config["loader_kwargs"]["organization"] = "HumanCompatibleAI"

        # Note: unfortunately we need to pass the venv **and** its name to the
        # huggingface policy loader since there is no way to get the name from the venv.
        # The name is needed to deduce the repo id and load the correct huggingface
        # model.
        e_config["loader_kwargs"]["env_name"] = config["environment"]["gym_id"]

    # Note: this only serves the purpose to indicated that you need to specify the
    #   path for local policies. It makes the config more explicit.
    if (
        e_config["policy_type"] in ("ppo", "sac")
        and "path" not in e_config["loader_kwargs"]
    ):  # pragma: no cover
        e_config["loader_kwargs"]["path"] = None
    return e_config


@expert_ingredient.capture
def get_expert_policy(venv, policy_type, loader_kwargs):
    return serialize.load_policy(policy_type, venv, **loader_kwargs)



"""This ingredient provides BC algorithm instance.

It is either loaded from disk or constructed from scratch.
"""
import warnings
from typing import Optional, Sequence

import sacred
import torch as th
from stable_baselines3.common import vec_env

from imitation.algorithms import bc
from imitation.data import types
from imitation.scripts.ingredients import policy

bc_ingredient = sacred.Ingredient("bc", ingredients=[policy.policy_ingredient])


@bc_ingredient.config
def config():
    batch_size = 32
    l2_weight = 3e-5  # L2 regularization weight
    optimizer_cls = th.optim.Adam
    optimizer_kwargs = dict(
        lr=4e-4,
    )
    train_kwargs = dict(
        n_epochs=None,  # Number of BC epochs per DAgger training round
        n_batches=None,  # Number of BC batches per DAgger training round
        log_interval=500,  # Number of updates between Tensorboard/stdout logs
    )
    agent_path = None  # Path to serialized policy. If None, a new policy is created.

    locals()  # quieten flake8 unused variable warning


@bc_ingredient.capture
def make_bc(
    venv: vec_env.VecEnv,
    expert_trajs: Sequence[types.Trajectory],
    custom_logger,
    batch_size: int,
    l2_weight: float,
    optimizer_cls,
    optimizer_kwargs,
    _rnd,
) -> bc.BC:
    return bc.BC(
        observation_space=venv.observation_space,
        action_space=venv.action_space,
        policy=make_or_load_policy(venv),
        demonstrations=expert_trajs,
        custom_logger=custom_logger,
        rng=_rnd,
        batch_size=batch_size,
        l2_weight=l2_weight,
        optimizer_cls=optimizer_cls,
        optimizer_kwargs=optimizer_kwargs,
    )


@bc_ingredient.capture
def make_or_load_policy(venv: vec_env.VecEnv, agent_path: Optional[str]):
    """Makes a policy or loads a policy from a path if provided.

    Args:
        venv: Vectorized environment we will be imitating demos from.
        agent_path: Path to serialized policy. If provided, then load the
            policy from this path. Otherwise, make a new policy.
            Specify only if policy_cls and policy_kwargs are not specified.

    Returns:
        A Stable Baselines3 policy.
    """
    if agent_path is None:
        policy.make_policy(venv)
    else:
        warnings.warn(
            "When agent_path is specified, policy.policy_cls and policy.policy_kwargs "
            "are ignored.",
            RuntimeWarning,
        )
        return bc.reconstruct_policy(agent_path)



"""Common wrapper for adding custom reward values to an environment."""

import collections
from typing import Deque

import numpy as np
from stable_baselines3.common import callbacks
from stable_baselines3.common import logger as sb_logger
from stable_baselines3.common import vec_env

from imitation.data import types
from imitation.rewards import reward_function


class WrappedRewardCallback(callbacks.BaseCallback):
    """Logs mean wrapped reward as part of RL (or other) training."""

    def __init__(self, episode_rewards: Deque[float], *args, **kwargs):
        """Builds WrappedRewardCallback.

        Args:
            episode_rewards: A queue that episode rewards will be placed into.
            *args: Passed through to `callbacks.BaseCallback`.
            **kwargs: Passed through to `callbacks.BaseCallback`.
        """
        self.episode_rewards = episode_rewards
        super().__init__(*args, **kwargs)

    def _on_step(self) -> bool:
        return True

    def _on_rollout_start(self) -> None:
        if len(self.episode_rewards) == 0:
            return
        mean = sum(self.episode_rewards) / len(self.episode_rewards)
        assert isinstance(self.logger, sb_logger.Logger)
        self.logger.record("rollout/ep_rew_wrapped_mean", mean)


class RewardVecEnvWrapper(vec_env.VecEnvWrapper):
    """Uses a provided reward_fn to replace the reward function returned by `step()`.

    Automatically resets the inner VecEnv upon initialization. A tricky part
    about this class is keeping track of the most recent observation from each
    environment.

    Will also include the previous reward given by the inner VecEnv in the
    returned info dict under the `original_env_rew` key.
    """

    def __init__(
        self,
        venv: vec_env.VecEnv,
        reward_fn: reward_function.RewardFn,
        ep_history: int = 100,
    ):
        """Builds RewardVecEnvWrapper.

        Args:
            venv: The VecEnv to wrap.
            reward_fn: A function that wraps takes in vectorized transitions
                (obs, act, next_obs) a vector of episode timesteps, and returns a
                vector of rewards.
            ep_history: The number of episode rewards to retain for computing
                mean reward.
        """
        assert not isinstance(venv, RewardVecEnvWrapper)
        super().__init__(venv)
        self.episode_rewards: Deque = collections.deque(maxlen=ep_history)
        self._cumulative_rew = np.zeros((venv.num_envs,))
        self.reward_fn = reward_fn
        self._old_obs = None
        self._actions = None
        self.reset()

    def make_log_callback(self) -> WrappedRewardCallback:
        """Creates `WrappedRewardCallback` connected to this `RewardVecEnvWrapper`."""
        return WrappedRewardCallback(self.episode_rewards)

    @property
    def envs(self):
        return self.venv.envs

    def reset(self):
        self._old_obs = self.venv.reset()
        return self._old_obs

    def step_async(self, actions):
        self._actions = actions
        return self.venv.step_async(actions)

    def step_wait(self):
        obs, old_rews, dones, infos = self.venv.step_wait()

        # The vecenvs automatically reset the underlying environments once they
        # encounter a `done`, in which case the last observation corresponding to
        # the `done` is dropped. We're going to pull it back out of the info dict!
        obs_fixed = []
        obs = types.maybe_wrap_in_dictobs(obs)
        for single_obs, single_done, single_infos in zip(obs, dones, infos):
            if single_done:
                single_obs = single_infos["terminal_observation"]

            obs_fixed.append(types.maybe_wrap_in_dictobs(single_obs))
        obs_fixed = (
            types.DictObs.stack(obs_fixed)
            if isinstance(obs, types.DictObs)
            else np.stack(obs_fixed)
        )
        rews = self.reward_fn(
            self._old_obs,
            self._actions,
            types.maybe_unwrap_dictobs(obs_fixed),
            np.array(dones),
        )
        assert len(rews) == len(obs), "must return one rew for each env"
        done_mask = np.asarray(dones, dtype="bool").reshape((len(dones),))

        # Update statistics
        self._cumulative_rew += rews
        for single_done, single_ep_rew in zip(dones, self._cumulative_rew):
            if single_done:
                self.episode_rewards.append(single_ep_rew)
        self._cumulative_rew[done_mask] = 0

        # we can just use obs instead of obs_fixed because on the next iteration
        # after a reset we DO want to access the first observation of the new
        # trajectory, not the last observation of the old trajectory
        obs = types.maybe_unwrap_dictobs(obs)
        self._old_obs = obs
        for info_dict, old_rew in zip(infos, old_rews):
            info_dict["original_env_rew"] = old_rew
        return obs, rews, dones, infos
